/*

Guillermo Leal Llaguno
www.evvisual.com
Corona to Vray Scene Converter.
tested in 3Dsmax 2024, corona 11 and vray 6 update 2.1

Based on the one that comes with Vray 6. 
The original script did not work correctly and the only other script available did not do things the way I wanted so I decided to modify the one that comes with vray 6.

It works for me but it might not work for you , I did things to fit my needs but if you find issues or have suggestions please let me know.

History:

Ver 1.05 2024/04/11
	- fix: fixed a crash caused by a parameter changed in corona 10+ for the AO texture.

Ver 1.04 2023/02/04
	- add: The cleanup slate editor seems to be working fine....
	- fix: few other fixes and cleanup
	- fix: set defualt of vraydistance Inside Solid to ON
	
Ver 1.03 2023/02/03
	- fix: removed all the _vray prefix to every map, texture, material, etc.  I want to keep the original names
	
Ver 1.02  - 2023/01/30:
	- add: Option to not convert the Corona Color Correct map, it turns out vray actually support it if you are using CPU, for GPU you do need to convert it and the result is not 100% correct.
	- fix: Render Elements where off after converting
	- add: if we had a denosier enabled then we add a VRayDenoiser Render Element
	- add: added a function to cleanup the slate editor of unused maps, and removed what I had added to do the same on each function but wasnt really working

Ver 1.01  - 2023/01/29:
	- fix/add: reworked the gt_coronaProxiesToVRay function. they were not being created as instances, also now you have the option to use existing proxies, for example if you already converted the scene and dont want to re-export the animated 
	proxies which take a lot of time.
	- fix: I was missing a forward declaration for (fn gt_createVRayMtl=undefined)
	
Ver 1.0   - 2023/01/28:
	- add: Support for converting from CoronaMultiMap to VRayMultiSubTex
	- add: Export Animated Proxies
	- fix: Export all proxies, the original missed a bunch which were copies and not instances
	- add: new GUI
	- fix: Several random tweaks
	- add: worked on the delete slate node function but it still not working 100% correct, its disabled for now.
	- add: all the maps conversion in gt_createVRayTexture  now have their own function, even the simple ones, so I can cleanup the nodes after conversion
	- add: added gt_converfFrom_Unsupported inTexture, which just adds a color for easy identifying, it needs more work
	- add: worked a bit on the map CoronaFrontBack, if there is a map in the front slot it returns that.  It needs more work, I know..
	- add: converts corona render elements to vray similar ones.  It needs more work; some just get deleted. Just did what I need it at the time.
	- fix: corona cameras with no target were crashing as it was asking for the target distance. it now checks if its a target camera
	- add: transfer material ID to the new material
	

Know Issues :
- The slate editor cleanup function needs more testing, it might delete things that you dont want to.
- some proxies get huge, need to find out why, I think they have negative transform
- Not sure why when using the existing proxy, the pivot of some proxy files is off.

TODO:
- DONE: avoid naming everything with the suffix Vray, I want everything to have the same name as before.  Proxies, Lights, Materials, etc
- DONE: Environment Maps are not being converted
- DONE: render elements get turn off on conversion
- AO texture did not come trough
- DONE: option to leave the coronacolor correct map, it is supported


*/
(
	Global gui_CoronaToVrayConverter	
	try(destroydialog gui_CoronaToVrayConverter)catch()
	
	---------------------------------------------------------------------------------------------------
	
	Struct Convert_Scene_CoronatoVray
	(
		version = "1.05",
		debug = true,
		
		-- enviormentMaps
		theEnviormentMap = undefined,
		theEnviormentColor = undefined,
		
		theCoronaSingleMap = undefined,
		theCoronaSingleColor = undefined,
		theCoronaSingleEnabled = undefined,
		theCoronaSingleEnabledTex = undefined,
		
		theCoronaVisibilityOverride = undefined,
		theCoronaVisibilityColor = undefined,
		theCoronaVisibilityEnabled =undefined,
		theCoronaVisibilityEnabledTex =undefined,
		
		theCoronaReflectOverride = undefined,
		theCoronaReflectColor = undefined,
		theCoronaReflectEnabled =  undefined,
		theCoronaReflectEnabledTex =  undefined,
		
		theCoronaRefractOverride= undefined,
		theCoronaRefractColor= undefined,
		theCoronaRefractEnabled = undefined,
		theCoronaRefractEnabledTex = undefined,
		
		-- used to look up materials that have already been converted
		glsc_orig_mtls=#(),
		glsc_new_mtls=#(),

		-- used to look up textures that have already been converted
		glsc_orig_texmaps=#(),
		glsc_new_texmaps=#(),

		-- used to look up bitmap textures that have already been converted
		glsc_orig_bitmaptextures=#(),
		glsc_new_hdritexmaps=#(),
		
		glsc_listOfConvertedNodesHandles=#(),
		
		proxyCustomPath = undefined,
		
		replaceImgCounter=0,
		replaceImgNums=0,	
		
		-- Bitmap2VRayBitmap converter
		convertB2HDRI_ConvertEXR=false,
		convertB2HDRI_ConverterErrors=0,
		convertB2HDRI_ErrorsInTXExisting=0, --named only for Tx because of upcoming exr conversion elimination

		listOfSuppFormats = #(".png", ".bmp", ".tga", ".hdr", ".sgi", ".rgb", ".rgba", ".jpg", ".jpeg", ".jpe", ".exr", ".pic", ".tif", ".tiff", ".psd"),
		mapfiles=#(), --list of bitmap image file paths	
		
		unsupportedColor = red,

		-- *********************************************************************** Helper Functions *************************************************************************

		fn isVRayCurrent = matchPattern (renderers.current as string) pattern:"*V_Ray*",
		
		fn isCoronaCurrent = matchPattern (renderers.current as string) pattern:"*Corona*",
		
		fn isCorona10orHiger =
			(
				if (isCoronaCurrent()) then
					(
						if CoronaRenderer.getVersionMajorNumber() >= 10 then
							return true
						else
							return false
					)else
					return false
			),
	
		fn gt_clamp val minVal maxVal  =
			(
				if val < minVal then val = minVal
				else if val > maxVal then val = maxVal
				return val
			),

		fn gt_remapValue oldVal oldMin oldMax newMin newMax =
			(
				oldRange = oldMax - oldMin
				newRange = newMax - newMin
				newVal = (oldVal / oldRange) * newRange
			),

		----Disney Specular to plain IoR
		fn gt_getIorFromDisney specularValue =
			(
				/*
				const float f0      = max(0.f, specularValue * 0.08f);
				const float divisor = 1.f - sqrt(f0);
				if (divisor <= 0.f) {
					return 3.f;
				}
				return gt_clamp(2.f / divisor - 1.f, 1.f, 3.f);
				*/

				f0 = amax #(0.0 , specularValue * 0.08)
				divisor = 1.0 - sqrt(f0)
				outIoR = if divisor <= 0.0 then 3.0	else (gt_clamp (2.0 / divisor - 1.0) 1.0 3.0)
			),

		fn gt_setupComplexIoR cpm=
			(
				theMap = VRayPluginNodeTex name:"complexIoRTex" plugin:"TexComplexFresnel"
				theMap.n_red = cpm.complexIorNRed
				theMap.n_green = cpm.complexIorNGreen
				theMap.n_blue = cpm.complexIorNBlue
				theMap.k_red = cpm.complexIorKRed
				theMap.k_green = cpm.complexIorKGreen
				theMap.k_blue = cpm.complexIorKBlue

				theMap
			),
			
		--here input param : 'ver' have to be presented as <number>.<number>.{<number>}
		fn gt_checkForPreviousCoronaVersion ver = 
			(
				res = false
				if CoronaRenderer != undefined do (
					local currVer = CoronaRenderer.getCoronaVersion()
					local parsedCurrVer = filterstring currVer " ,."
					local intCurrVer = ((parsedCurrVer[1]+parsedCurrVer[2]) as integer)
					local parsedVer = filterstring ver " ,."
					local intVer = ((parsedVer[1]+parsedVer[2]) as integer)
					if intCurrVer != undefined and intVer != undefined then
						res = (intCurrVer < intVer)
				)
				return res
			),
			
		-- Function to see if a map is used in the scene
		fn isMapUsedInTheScene theMap =
		(
			theRef = refs.dependents theMap
			
			for a = 1 to theRef.count do			
			(
	-- 			format "class: % : % \n" (classof theRef[a]) theRef[a]
				if ( (theRef[a] != rootScene) and (theRef[a] != meditMaterials ) )then (
					return true
				)
			)

			return false
			
		),

		-- *********************************************************************** Render Element Functions **************************************************************************
			
	
		-- Function to print all the information about all exiting elements in the scene
		fn printInfoOnRenderElements =
			(
				remCurrent = maxOps.GetCurRenderElementMgr()

				for i = 0 to remCurrent.NumRenderElements()-1 do --note render element manager is 0 based
				(
					re = remCurrent.GetRenderElement i
				
					format "Element Class: %\n" re
					format " Name: % \n" re.elementname
					format "  Filename: %" (remCurrent.GetRenderElementFileName i)
					format "  %\n\n" (show re)
					
					-- clear the render element filename
					remCurrent.SetRenderElementFilename i ""
				)
			),

		-- function to convert the render elements from corona to vray.
		-- not all elements have a conversion, some just geet deleted for now. I might come later and add the code
		fn convert_RenderElements_from_CoronaToVray deleteOld:true debug:false =
			(
				eleMgr = maxOps.GetCurRenderElementMgr()

				if debug then
					format "Converting Render Elements... \n"
					
				for i = 0 to eleMgr.NumRenderElements()-1 do --note render element manager is 0 based
				(
					ele = eleMgr.GetRenderElement i
					
					if debug then
						format "Converting Element: % : % \n"  (classof ele) ele.elementname

					case of 
						(
							((classof ele) == CESSENTIAL_Direct):
								(
									theName = ele.elementname
									isOn = ele.enabled
									filterOn = ele.filter
									doDenoise = ele.useDenoise

									ele = VRayLighting enabled:true color_mapping:true multiplier:1.0
									ele.elementName = theName
									ele.enabled = isOn
									ele.filterOn = filterOn
									ele.vrayVFB = true
									ele.color_mapping = true
									ele.multiplier = 1.0
									ele.denoise = doDenoise
									
									eleMgr.addRenderElement ele
								)
						
							((classof ele) == CESSENTIAL_Indirect):
								(
									theName = ele.elementname
									isOn = ele.enabled
									filterOn = ele.filter
									doDenoise = ele.useDenoise
									
									ele = VRayGlobalIllumination enabled:true color_mapping:true multiplier:1.0
									ele.elementName = theName
									ele.enabled = isOn
									ele.filterOn = filterOn
									ele.vrayVFB = true
									ele.color_mapping = true
									ele.multiplier = 1.0
									ele.denoise = doDenoise
									
									eleMgr.addRenderElement ele
								)
									
							((classof ele) == CESSENTIAL_Emission):
								(
									theName = ele.elementname
									isOn = ele.enabled
									filterOn = ele.filter

									ele = VRaySelfIllumination enabled:true color_mapping:true multiplier:1.0
									ele.elementName = theName
									ele.enabled = isOn
									ele.filterOn = filterOn
									ele.vrayVFB = true
									ele.color_mapping = true
									ele.multiplier = 1.0
									
									eleMgr.addRenderElement ele
								)

							((classof ele) == CESSENTIAL_Reflect):
								(
									theName = ele.elementname
									isOn = ele.enabled
									filterOn = ele.filter							
									doDenoise = ele.useDenoise

									ele = VRayReflection enabled:true color_mapping:true multiplier:1.0
									ele.elementName = theName
									ele.enabled = isOn
									ele.filterOn = filterOn
									ele.vrayVFB = true
									ele.color_mapping = true
									ele.multiplier = 1.0
									ele.denoise = doDenoise
									
									eleMgr.addRenderElement ele
								)

							((classof ele) == CESSENTIAL_Refract):
								(
									theName = ele.elementname
									isOn = ele.enabled
									filterOn = ele.filter	
									doDenoise = ele.useDenoise							
									
									ele = VRayRefraction enabled:true color_mapping:true multiplier:1.0
									ele.elementName = theName
									ele.enabled = isOn
									ele.filterOn = filterOn
									ele.vrayVFB = true
									ele.color_mapping = true
									ele.multiplier = 1.0
									ele.denoise = doDenoise
									
									eleMgr.addRenderElement ele
								)
										
							((classof ele) == CESSENTIAL_Translucency):
								(
								)
									
							((classof ele) == CESSENTIAL_Volumetrics):
								(
									theName = ele.elementname
									isOn = ele.enabled
									filterOn = ele.filter	
									doDenoise = ele.useDenoise
									
									ele = VRayAtmosphere enabled:true color_mapping:true multiplier:1.0
									ele.elementName = theName
									ele.enabled = isOn
									ele.filterOn = filterOn
									ele.vrayVFB = true
									ele.color_mapping = true
									ele.multiplier = 1.0
									ele.denoise = doDenoise
									
									eleMgr.addRenderElement ele
								)

							((classof ele) == CGeometry_NormalsDotProduct):
								(
								)

							((classof ele) == CGeometry_NormalsGeometry):
								(
								)

							((classof ele) == CGeometry_NormalsShading):
								(
									theName = ele.elementname
									isOn = ele.enabled
									filterOn = ele.filter							
									
									ele = VRayNormals enabled:true 
									ele.elementName = theName
									ele.enabled = isOn
									ele.filterOn = filterOn
									ele.vrayVFB = true
									
									eleMgr.addRenderElement ele
								)

							((classof ele) == CGeometry_UvwCoords):
								(
								)

							((classof ele) == CGeometry_Velocity):
								(
									theName = ele.elementname
									isOn = ele.enabled
									filterOn = ele.filter	

									ele = VRayVelocity enabled:true c
									ele.elementName = theName
									ele.enabled = isOn
									ele.filterOn = filterOn
									ele.vrayVFB = true
									
									eleMgr.addRenderElement ele
								)

							((classof ele) == CGeometry_WorldPosition):
								(
								)

							((classof ele) == CGeometry_ZDepth):
								(
									theName = ele.elementname
									isOn = ele.enabled
									filterOn = ele.filter	
									
									zMin = ele.minZ
									zMax = ele.maxZ
									useCClip = ele.useCameraEnvironmentRange

									ele = VRayZDepth enabled:true 
									ele.elementName = theName
									ele.enabled = isOn
									ele.filterOn = filterOn
									ele.vrayVFB = true
									
									ele.zdepth_use_camera_clip_boundaries = useCClip
									ele.zdepth_min = zMin
									ele.zdepth_max = zMax
									
									eleMgr.addRenderElement ele
								)

							((classof ele) == CInfo_NetworkRenderDebugging):
								(
								)

							((classof ele) == CInfo_RenderStamp):
								(
								)

							((classof ele) == CInfo_SamplingFocus):
								(
								)

							((classof ele) == CInfo_TimePerSample):
								(
								)

							((classof ele) == CMasking_Cryptomatte):
								(
									theName = ele.elementname
									isOn = ele.enabled
									filterOn = ele.filter	
									
									inLevels = ele.levels
									inIDTyle = ele.idType						
									
									ele = VRayCryptomatte enabled:true 
									ele.elementName = theName
									ele.enabled = isOn
									ele.filterOn = filterOn
									ele.vrayVFB = true
									
									if inIDTyle == 2 then
										inIDTyle = 4
									ele.id_type = inIDTyle
									ele.num_levels = inLevels
		  
									eleMgr.addRenderElement ele
								)

							((classof ele) == CMasking_ID):
								(
								)

							((classof ele) == CMasking_Mask):
								(
									IN_objectR = ele.objectR
									IN_objectROn = ele.objectROn 
									IN_objectG = ele.objectG 
									IN_objectGOn =  ele.objectGOn 
									IN_objectB =  ele.objectB
									IN_objectBOn = ele.objectBOn

									IN_materialR = ele.materialR 
									IN_materialROn = ele.materialROn
									IN_materialG = ele.materialG 
									IN_materialGOn = ele.materialGOn 
									IN_materialB =  ele.materialB 
									IN_materialBOn = ele.materialBOn 

									IN_nodesR = ele.nodesR 
									IN_nodesROn = ele.nodesROn 
									IN_nodesRIncludeMod = ele.nodesRIncludeMod 
									IN_nodesG = ele.nodesG 
									IN_nodesGOn = ele.nodesGOn
									IN_nodesGIncludeMod = ele.nodesGIncludeMod 
									IN_nodesB = ele.nodesB
									IN_nodesBOn = ele.nodesBOn 
									IN_nodesBIncludeMod = ele.nodesBIncludeMod 

									theName = ele.elementname
									isOn = ele.enabled
									filterOn = ele.filter

									ele = MultiMatteElement enabled:true
									ele.elementName = theName
									ele.enabled = isOn
									ele.filterOn = filterOn
									ele.vrayVFB = true

									if IN_objectROn then (
										ele.R_gbufIDOn = true
										ele.R_gbufID = IN_objectR
										)

									if IN_objectGOn then (
										ele.G_gbufIDOn = true
										ele.G_gbufID = IN_objectG
										)

									if IN_objectBOn then (
										ele.B_gbufIDOn = true
										ele.B_gbufID = IN_objectB
										)

									if IN_materialROn or IN_materialGOn or IN_materialBOn then
											ele.MatID = true
										else
											ele.MatID = false

									if IN_materialROn then (
										ele.R_gbufID = IN_materialR
										)

									if IN_materialGOn then (
										ele.G_gbufID = IN_materialG
										)

									if IN_materialBOn then (
										ele.B_gbufID = IN_materialB
										)

									if IN_nodesROn then
										ele.redList = IN_nodesR
									if IN_nodesGOn then
										ele.greenList = IN_nodesG
									if IN_nodesBOn then
										ele.blueList = IN_nodesB

									ele.redListType = IN_nodesRIncludeMod
									ele.greenListType = IN_nodesGIncludeMod
									ele.blueListType = IN_nodesBIncludeMod

									eleMgr.addRenderElement ele	
								)
								
							((classof ele) == CMasking_WireColor):
								(
									theName = ele.elementname
									isOn = ele.enabled
									filterOn = ele.filter							
									
									ele = VRayWireColor enabled:true 
									ele.elementName = theName
									ele.enabled = isOn
									ele.filterOn = filterOn
									ele.vrayVFB = true
									
									eleMgr.addRenderElement ele
								)

							((classof ele) == CShading_Albedo):
								(
								)

							((classof ele) == CShading_Alpha):
								(
									theName = ele.elementname
									isOn = ele.enabled
									filterOn = ele.filter						
									
									ele = VRayAlpha enabled:true 
									ele.elementName = theName
									ele.enabled = isOn
									ele.filterOn = filterOn
									
									eleMgr.addRenderElement ele
								)

							((classof ele) == CShading_Beauty):
								(
								)																				

							((classof ele) == CShading_BloomGlare):
								(
								)

							((classof ele) == CShading_Caustics):
								(	
									theName = ele.elementname
									isOn = ele.enabled
									filterOn = ele.filter	
									doDenoise = ele.denoiseBlend						
									
									ele = VRayCaustics enabled:true 
									ele.elementName = theName
									ele.enabled = isOn
									ele.filterOn = filterOn
									ele.vrayVFB = true
									ele.color_mapping = true
									ele.multiplier = 1.0
									ele.denoise = doDenoise
									
									eleMgr.addRenderElement ele
								)

							((classof ele) == CShading_Components):
								(
								)

							((classof ele) == CShading_LightMix):
								(
									theName = ele.elementname
									isOn = ele.enabled
									filterOn = ele.filter							
									
									ele = VRayLightMix enabled:true 
									ele.elementName = theName
									ele.enabled = isOn
									ele.filterOn = filterOn
									ele.vrayVFB = true
									ele.denoise = true
									
									ele.create_rest_channel = true
									ele.group_by = 4 -- Manual Light Select
									
									eleMgr.addRenderElement ele
								)

							((classof ele) == CShading_LightSelect):
								(			
									theName = ele.elementname
									isOn = ele.enabled
									filterOn = ele.filter	
									doDenoise = ele.useDenoise
									
									tmpItems = ele.includedNodes	
									In_includedNodes = #()
									
									-- the corona lightselect support a mix of lights and mesh with cornalight material, vray does not. so here we remove the geometry
									for t = 1 to tmpItems.count do (
										if superclassof tmpItems[t] == light then
											append In_includedNodes tmpItems[t]
									)						
									
									ele = VRayLightSelect enabled:true color_mapping:true multiplier:1.0
									ele.elementName = theName
									ele.enabled = isOn
									ele.filterOn = filterOn
									ele.vrayVFB = true
									ele.color_mapping = true
									ele.multiplier = 1.0
									ele.denoise = doDenoise
									
									ele.lights_exclude_type = 1 -- Include
									ele.list_exclude_granularity = 0 -- Lights
									ele.mode = 4 -- Full
									ele.lights = In_includedNodes
									
									eleMgr.addRenderElement ele
								)

							((classof ele) == CShading_RawComponent):
								(							
								)

							((classof ele) == CShading_Shadows):
								(	
									theName = ele.elementname
									isOn = ele.enabled
									filterOn = ele.filter	
									doDenoise = ele.useDenoise							
									
									ele = VRayShadows enabled:true color_mapping:true multiplier:1.0
									ele.elementName = theName
									ele.enabled = isOn
									ele.filterOn = filterOn
									ele.vrayVFB = true
									ele.color_mapping = true
									ele.multiplier = 1.0
									ele.denoise = doDenoise
									
									eleMgr.addRenderElement ele
								)

							((classof ele) == CShading_SourceColor):
								(
								)

							((classof ele) == CTexmap):
								(			
								theName = ele.elementname
								isOn = ele.enabled
								filterOn = ele.filter	
								doDenoise = ele.useDenoise
									
								In_texmap = ele.texmap
								In_excludedObjects = ele.excludedObjects
								In_overrideBackgroundColor = ele.overrideBackgroundColor
									
								ele = VRayExtraTex enabled:true 
								ele.elementName = theName
								ele.enabled = isOn
								ele.filterOn = filterOn
								ele.vrayVFB = true
								ele.denoise = doDenoise
									
								ele.texture = In_texmap
								ele.excludeList = In_excludedObjects
								ele.background_color = In_overrideBackgroundColor
									
								eleMgr.addRenderElement ele
								)				

							default: ()
						) -- end case
				) -- end for loop
				
				if debug then
					format "\n" 
					
				-- if we are deleting the old elements it has to be after we added the new ones as not to change the element count.
				if deleteOld then
				(
					for i =  (eleMgr.NumRenderElements()-1) to 0 by -1 do 
						(
							ele = eleMgr.GetRenderElement i

							if debug then
								format "Deleting Element: % : % \n"  (classof ele) ele.elementname

							case of 
								(
									((classof ele) == CESSENTIAL_Direct):
										(
												eleMgr.RemoveRenderElement ele	
										)

									((classof ele) == CESSENTIAL_Indirect):
										(
												eleMgr.RemoveRenderElement ele	
										)

									((classof ele) == CESSENTIAL_Emission):
										(
												eleMgr.RemoveRenderElement ele							
										)

									((classof ele) == CESSENTIAL_Reflect):
										(
												eleMgr.RemoveRenderElement ele							
										)

									((classof ele) == CESSENTIAL_Refract):
										(
												eleMgr.RemoveRenderElement ele							
										)

									((classof ele) == CESSENTIAL_Translucency):
										(
												eleMgr.RemoveRenderElement ele
										)

									((classof ele) == CESSENTIAL_Volumetrics):
										(
												eleMgr.RemoveRenderElement ele							
										)

									((classof ele) == CGeometry_NormalsDotProduct):
										(
												eleMgr.RemoveRenderElement ele
										)

									((classof ele) == CGeometry_NormalsGeometry):
										(
												eleMgr.RemoveRenderElement ele
										)
														
									((classof ele) == CGeometry_NormalsShading):
										(
												eleMgr.RemoveRenderElement ele							
										)

									((classof ele) == CGeometry_UvwCoords):
										(
												eleMgr.RemoveRenderElement ele
										)

									((classof ele) == CGeometry_Velocity):
										(
												eleMgr.RemoveRenderElement ele							
										)

									((classof ele) == CGeometry_WorldPosition):
										(
												eleMgr.RemoveRenderElement ele
										)
																	
									((classof ele) == CGeometry_ZDepth):
										(
												eleMgr.RemoveRenderElement ele							
										)

									((classof ele) == CInfo_NetworkRenderDebugging):
										(
												eleMgr.RemoveRenderElement ele
										)

									((classof ele) == CInfo_RenderStamp):
										(
												eleMgr.RemoveRenderElement ele
										)

									((classof ele) == CInfo_SamplingFocus):
										(
												eleMgr.RemoveRenderElement ele
										)

									((classof ele) == CInfo_TimePerSample):
										(
												eleMgr.RemoveRenderElement ele
										)

									((classof ele) == CMasking_Cryptomatte):
										(
												eleMgr.RemoveRenderElement ele							
										)

									((classof ele) == CMasking_ID):
										(
												eleMgr.RemoveRenderElement ele
										)

									((classof ele) == CMasking_Mask):
										(
												eleMgr.RemoveRenderElement ele							
										)

									((classof ele) == CMasking_WireColor):
										(
												eleMgr.RemoveRenderElement ele							
										)

									((classof ele) == CShading_Albedo):
										(
												eleMgr.RemoveRenderElement ele
										)

									((classof ele) == CShading_Alpha):
										(
												eleMgr.RemoveRenderElement ele							
										)

									((classof ele) == CShading_Beauty):
										(
												eleMgr.RemoveRenderElement ele
										)																				

									((classof ele) == CShading_BloomGlare):
										(
												eleMgr.RemoveRenderElement ele
										)

									((classof ele) == CShading_Caustics):
										(
												eleMgr.RemoveRenderElement ele							
										)

									((classof ele) == CShading_Components):
										(
												eleMgr.RemoveRenderElement ele
										)

									((classof ele) == CShading_LightMix):
										(
												eleMgr.RemoveRenderElement ele							
										)

									((classof ele) == CShading_LightSelect):
										(
												eleMgr.RemoveRenderElement ele							
										)

									((classof ele) == CShading_RawComponent):
										(
												eleMgr.RemoveRenderElement ele
										)

									((classof ele) == CShading_Shadows):
										(
												eleMgr.RemoveRenderElement ele							
										)

									((classof ele) == CShading_SourceColor):
										(
												eleMgr.RemoveRenderElement ele
										)

									((classof ele) == CTexmap):
										(
												eleMgr.RemoveRenderElement ele							
										)				

									default: ()
								) -- end case

						) -- end for loop
				)
			
				-- if we have some type of denoising enable then add the denois render Element
				if renderers.current.denoise_filterType != 0 then
				(
					ele = VRayDenoiser enabled:true 
					ele.elementName = "VRayDenoiser"
					ele.enabled = true
					ele.filterOn = true
					ele.vrayVFB = true
					eleMgr.addRenderElement ele
				)
				
				-- make sure they are on
				eleMgr.SetElementsActive true
				
				if debug then
					format "\nFinished Converting Render Elements \n"
				
			),
	
		-- *********************************************************************** Color Functions **************************************************************************		
		fn gt_vrayRGBtoY inputmap type =
			(
			-- RGB intensity as ITU BT.709 Y = 0.2126 R + 0.7152 G + 0.0722 B
			-- If not then use ITU BT.601 Y = 0.299 R + 0.587 G + 0.114 B
			rgbColors=#(#(([0.2126,0.2126,0.2126]*255) as color,([0.7152,0.7152,0.7152]*255) as color,([0.0722,0.0722,0.0722]*255) as color),
						#(([0.299,0.299,0.299]*255) as color,([0.587,0.587,0.587]*255) as color,([0.114,0.114,0.114]*255) as color))
			theRGBTint = RGBTint name:"RGB to Intensity" map1:inputmap
			if type == "BT.709" then
			(
				theRGBTint.Red = rgbColors[1][1]
				theRGBTint.Green = rgbColors[1][2]
				theRGBTint.Blue = rgbColors[1][3]
			)
			else if type == "BT.601" do
			(
				theRGBTint.Red = rgbColors[2][1]
				theRGBTint.Green = rgbColors[2][2]
				theRGBTint.Blue = rgbColors[2][3]
			)
			theRGBTint
			),

		fn gt_vrayMultiplyRGBWeight rgbMap weightMap type= -- type is either "BT.709" or "BT.601"
			(
			theCompTex = VRayCompTex name:"Weight Multiplier"
			theCompTex.sourceA = gt_vrayRGBtoY weightMap type
			theCompTex.sourceB = rgbMap
			theCompTex.operator = 3 -- Multiply
			theCompTex
			),

		-- This function uses the Max Output map (Or a VRayCompTex, and a color, to do 1-texture) to invert the incoming texture
		-- Only if it's not a bitmap (i.e. doesn't have its own .output section), or if it is instanced.
		fn gt_vrayInvertTexture rgbMap doVRayOnly=
			(
			rgbMapClass=classof rgbMap
			if rgbMapClass != Bitmaptexture and rgbMapClass != VRayBitmap or (refhierarchy.IsRefTargetInstanced rgbMap) then
			(
				if doVRayOnly then
				(
					theCompTex=VRayCompTex name:"Inverter"
					theCompTex.sourceA = vrayColor color:white name:"One"
					theCompTex.sourceB = rgbMap
					theCompTex.operator = 1 --subtract (A-B)
					theOutput = theCompTex
				)
				else
				(
					theOutput=Output name:("Invert " + rgbMap.name) map1:rgbMap
					theOutput.output.invert = true
				)
			)
			else -- If it's a non-instanced bitmap, it's safe to toggle its invert flag.
			(
				rgbMap.output.invert = not rgbMap.output.invert --toggle, do not set!
				theOutput = rgbMap
			)
			theOutput
			),
			
		fn gt_normalizeColor col = 
			(
				norm=if (col.r>col.g) then (if (col.r>col.b) then col.r else col.b) else (if (col.g>col.b) then col.g else col.b)
				result=color ((col.r*255.0)/norm) ((col.g*255.0)/norm) ((col.b*255.0)/norm)
				result
			),			
			
			
		-- *********************************************************************** Shadows Functions **************************************************************************
			
		fn gt_shadowsNode node = 
			(
				undo off
				(
					if (isValidNode node) then (
						scls=superclassof node
						if (scls==light) then (
							try (
								olds=node.shadowGenerator
								cls=classof olds
								if (cls!=shadowMap and cls!=VRayShadow) then (
									if debug do
										format "  %\n" node
									news=VRayShadow()
									node.shadowGenerator=news
								)
							)
							catch ()
						)
					)

					for i=1 to node.children.count do (
						gt_shadowsNode node.children[i]
					)
				)
			),

		-- convert all node shadow types except shadow maps to VRay shadows
		fn gt_shadowsToVRay node = 
			(
				if debug do
					format "Converting node % shadows...\n" node.name
				gt_shadowsNode node
			-- 		format "Done\n"
			),	
			
		-- *********************************************************************** Lights Functions ****************************************************************************		

		fn gt_convertFrom_CoronaIESLight cLt =
			(
				undo off
				(
					local newLight=CreateInstance VRayIES name:cLt.name

					newLight.enabled = cLt.on
					newLight.targeted = cLt.targeted

					case of (
						(cLt.excludeList != undefined and cLt.excludeIncludeMod == true): (
							newLight.includeList = cLt.excludeList
						)
						(cLt.excludeList != undefined and cLt.excludeIncludeMod == false): (
							newLight.excludeList = cLt.excludeList
						)
						default: ()
					)

					-- Intensitiy, Color, Texmap
					newLight.power = cLt.intensity

					newLight.color_mode = cLt.colorMode -- 0: direct color, 1: kelvin temp
					newLight.color = cLt.color
					if cLt.colorMode == 1 do newLight.color_temperature = cLt.blackbodyTemp -- kelvin

					-- shape   -- 0: sphere, 1: rectangle, 2: disk, 3: cylinder
					newLight.override_shape = on
					case cLt.shape of (
						0:( newLight.shape = 3 -- sphere
							newLight.diameter = cLt.width * 2
						)
						1:( newLight.shape = 1 -- rectangle
							newLight.width = cLt.width
							newLight.length = cLt.height
						)
						2:( newLight.shape = 2 -- circle
							newLight.diameter = cLt.width * 2
						)
						3:( newLight.shape = 4  -- vertical cylinder
							newLight.diameter = cLt.width * 2.0
							newLight.height = cLt.height
						)
						default:(
						)
					)

					-- visibility
					newLight.cast_shadows = cLt.occludeOtherLights

					-- IES
					if cLt.iesFile != undefined do(
						newLight.ies_file = cLt.iesFile
					)

					newLight
				)
			),
				
		fn gt_replaceCoronaLightNodeVRayLight cLt = 
			(
				undo off
				(
					local theVRL = undefined
					if cLt.iesOn == true and cLt.iesFile != undefined and cLt.iesFile != "" then (
						theVRL = gt_convertFrom_CoronaIESLight cLt
					) else (
						theVRL=CreateInstance VRayLight name:cLt.name on:cLt.on castShadows:cLt.occludeOtherLights invisible:(not cLt.visibleDirectly) \
						affect_reflections:cLt.visibleRefl color_mode:cLt.colorMode  color:cLt.color targeted:cLt.targeted

						if cLt.colorMode == 1 do theVRL.color_temperature = cLt.blackbodyTemp
						if cLt.colorMode == 2 do theVRL.texmap_on = true
						if cLt.texmap != undefined do theVRL.texmap = cLt.texmap

						if cLt.targeted == true do theVRL.target_distance = cLt.targetDistance

						theVRL.lightDistribution = cLt.directionality

						--Include/Exclude
						if (cLt.excludeList != undefined and cLt.excludeList.count !=0) do (
							theVRL.inclExclType = 1
							if cLt.excludeIncludeMod then theVRL.includeList = cLt.excludeList else theVRL.excludeList = cLt.excludeList
						)

						--Color/ Intensity
						theVRL.normalizeColor = case of
						(
							(cLt.intensityUnits == 0):0
							(cLt.intensityUnits == 1):1
							default:0
						)

						theVRL.multiplier = cLt.intensity

						--shape
						case cLt.shape  of
						(
							1: ( --Plane
								theVRL.type = 0
								theVRL.size0 = cLt.width/2.0
								theVRL.size1 = cLt.height/2.0
							)
							0: ( --Sphere
								theVRL.type = 2
								theVRL.size0 = cLt.width
							)
							2: ( --Disc
								theVRL.type = 4
								theVRL.size0 = cLt.width/2.0
							)
							3: ( --Cylinder
								theVRL.type = 2
								theVRL.size0 = cLt.width
							)
						)
					)
					instCount = InstanceMgr.GetInstances cLt &arrInst
					for i=1 to instCount do
						arrInst[i].name = (arrInst[i].name+"_VRayLight")
					replaceInstances cLt theVRL
					theVRL
				)
			),
		
		fn gt_replaceCoronaSunLightNodeVRayLight coronaSLNode = 
			(
				undo off
				(
					local sunObj = undefined
					local targetObj = undefined
					local previousVersion = gt_checkForPreviousCoronaVersion("1.2.1")

					if previousVersion then (
						sunObj = VRaySun name:coronaSLNode.name enabled:coronaSLNode.on intensity_multiplier:coronaSLNode.intensity size_multiplier:coronaSLNode.sizeMultiplier invisible:(not coronaSLNode.visibleDirectly)
						targetObj = dummy name:(coronaSLNode.name+".Target") pos:coronaSLNode.pos  --then target
						targetObj.boxsize = [2,2,2]
						targetObj.lookat = sunObj --this makes the target for the sunn out of dummy object
						sunObj.pos = coronaSLNode.pos + coronaSLNode.displaySize*[sin coronaSLNode.height, 0.0, cos coronaSLNode.height] -- xy
						sunObj.pos = [coronaSLNode.pos.x , coronaSLNode.pos.y, sunObj.pos.z]  + [cos coronaSLNode.sunRotation , sin coronaSLNode.sunRotation, 0.0]*sqrt(coronaSLNode.displaySize^2 - sunObj.pos.z^2) --z

						--animated sun
						if coronaSLNode.sunRotation.isAnimated == true or coronaSLNode.height.isAnimated == true do
						(
							animate on
							for t in animationRange.start to animationRange.end by 1 do
							(
								at time t
								(
									sunObj.pos = coronaSLNode.pos + coronaSLNode.displaySize*[sin coronaSLNode.height, 0.0, cos coronaSLNode.height]
									sunObj.pos = [coronaSLNode.pos.x , coronaSLNode.pos.y, sunObj.pos.z]  + [cos coronaSLNode.sunRotation , sin coronaSLNode.sunRotation, 0.0]*sqrt(coronaSLNode.displaySize^2 - sunObj.pos.z^2)
								)
							)
						)
					) else (
						sunObj = CreateInstance VRaySun name:coronaSLNode.name enabled:coronaSLNode.on intensity_multiplier:coronaSLNode.intensity size_multiplier:coronaSLNode.sizeMultiplier invisible:(not coronaSLNode.visibleDirectly)
					)

					if (coronaSLNode.colorMode == 0) do (
						sunObj.color_mode = 1
						sunObj.filter_Color = coronaSLNode.colorDirect
					)

					--excludeList
					if coronaSLNode.excludeList.count != 0 do
					(
						sunObj.inclExclType = 1
						if coronaSLNode.excludeListIncludeMod == false then sunObj.excludeList = coronaSLNode.excludeList else sunObj.includeList = coronaSLNode.excludeList
					)

					if previousVersion then (
						theLayer = layermanager.getLayerFromName coronaSLNode.layer.name
						theLayer.addNode sunObj
						theLayer.addNode targetObj

						delete coronaSLNode
					) else (
						instCount = InstanceMgr.GetInstances coronaSLNode &arrInst
						for i=1 to instCount do (
							arrInst[i].name = arrInst[i].name
						)
						replaceInstances coronaSLNode sunObj
					)

					return undefined
				)
			),	

		fn gt_replaceLightNodeVRIESLight node = 
			(
				undo off
				(
					local newLight=CreateInstance VRayIES name:node.name

					if (newLight!=undefined) do (
						newLight.targeted=(node.target!=undefined)
						newLight.enabled=node.on
						newLight.color_mode=0
						newLight.color=gt_normalizeColor node.color
						newLight.rotation_X=node.xRotation
						newLight.rotation_Y=node.yRotation
						newLight.rotation_Z=node.zRotation
						newLight.use_light_shape=1
						if (node.webFile!=undefined) then (
							newLight.ies_file=node.webFile
							local powerCoeff=node.intensity/node.originalIntensity
							newLight.power=newLight.power*powerCoeff
						) else (
							newLight.power=node.intensity*4*pi
						)
						if (node.useMultiplier) do (newLight.power=(newLight.power*node.multiplier)/100.0)
						if (hasProperty newLight "override_shape") then (
							newLight.override_shape=true
							case of (
								(node.type==#Target_Point or node.type==#Free_Point): (
									newLight.shape=0
								)

								(node.type==#Target_Line or node.type==#Free_Line or
								node.type==#Target_Linear or node.type==#Free_Linear): (
									newLight.shape=1
									newLight.length=node.light_length
								)

								(node.type==#Target_Rectangle or node.type==#Free_Rectangle or
								node.type==#Target_Area or node.type==#Free_Area): (
									newLight.shape=1
									newLight.width=node.light_width
									newLight.length=node.light_length
								)

								(node.type==#Target_Disc or node.type==#Free_Disc): (
									newLight.shape=2
									newLight.diameter=node.light_radius*2
								)

								(node.type==#Target_Sphere or node.type==#Free_Sphere): (
									newLight.shape=3
									newLight.diameter=node.light_radius*2
								)

								(node.type==#Target_Cylinder or node.type==#Free_Cylinder): (
									newLight.shape=5
									newLight.length=node.light_length
									newLight.diameter=node.light_radius*2
								)
							)
						)
						if (node.includeList!=undefined) do (newLight.includeList=node.includeList)
						if (node.excludeList!=undefined) do (newLight.excludeList=node.excludeList)
					)

					instCount = InstanceMgr.GetInstances node &arrInst
					for i=1 to instCount do (
						arrInst[i].name = arrInst[i].name
					)
					replaceInstances node newLight
					newLight
				)
			),

		fn gt_replaceLightNodeVRayLight node =
			(
				undo off
				(
					local newLight=CreateInstance VRayLight name:node.name

					if (newLight!=undefined) do (
						newLight.targeted=(node.target!=undefined)
						if newLight.targeted then newLight.target_distance = node.targetDistance
						newLight.on=node.on
						newLight.color_mode=0
						newLight.color=gt_normalizeColor node.color
						newLight.normalizeColor = 1
						newLight.multiplier=node.intensity*4*pi
						newLight.invisible= not node.baseObject.Area_Light_Sampling_Custom_Attribute.mr_EnableLightShapeRendering
						newLight.texmap_on=node.projector
						if (node.projector) do newLight.texmap=node.projectorMap
						if (node.useMultiplier) do (newLight.multiplier=(newLight.multiplier*node.multiplier)/100.0)
						case of (
							(node.type==#Target_Rectangle or node.type==#Free_Rectangle or
							node.type==#Target_Area or node.type==#Free_Area): (
								newLight.type=0
								newLight.size0=node.light_width / 2.0
								newLight.size1=node.light_length / 2.0
								newLight.DoubleSided = true
							)

							(node.type==#Target_Disc or node.type==#Free_Disc): (
								newLight.type=4
								newLight.size0=node.light_radius
								newLight.DoubleSided = true
							)

							(node.type==#Target_Sphere or node.type==#Free_Sphere): (
								newLight.type=2
								newLight.size0=node.light_radius
							)
						)

						if (node.includeList!=undefined) do (newLight.includeList=node.includeList)
						if (node.excludeList!=undefined) do (newLight.excludeList=node.excludeList)
					)

					instCount = InstanceMgr.GetInstances node &arrInst
					for i=1 to instCount do
						arrInst[i].name = arrInst[i].name
					replaceInstances node newLight
					newLight
				)
			),	
		
		fn gt_replaceLight_mr_Sun lightsys = 
			(
				undo off
				(
					if (classof lightsys.sun==mr_Sun) do (
						local mrSun_on=lightsys.sun.on
						local mrSun_skymult=lightsys.sun.skymult

						lightsys.sun=createInstance VRaySun
						lightsys.sun.enabled=mrSun_on
						lightsys.sun.intensity_multiplier=mrSun_skymult

						lightsys.sky=undefined
					)
				)
			),	
		
		fn gt_replaceLightNode node = 
			(
				undo off 
				(
					local lightClassStandard=#(Free_Light, Target_Light,
													Free_Point, Target_Point,
													Free_Linear, Target_Linear,
													Free_Area, Target_Area,
													Free_Disc, Target_Disc,
													Free_Sphere, Target_Sphere,
													Free_Cylinder, Target_Cylinder)

					local lightClassCorona=#(CoronaLight, CoronaSun)
					local newLight
					if (isValidNode node) do (
						if ((findItem lightClassStandard (classof node))!=0) do (
							case (node.distribution) of (
								0: newLight=gt_replaceLightNodeVRayLight node	-- Uniform Spherical
								1: newLight=undefined						-- Spotlight
								2: newLight=gt_replaceLightNodeVRayLight node	-- Uniform Diffuse
								3: newLight=gt_replaceLightNodeVRIESLight node	-- Photometric Web
							)
						)

						local foundCL = 0
						if ((foundCL = findItem lightClassCorona (classof node))!=0) do (
							case (foundCL) of (
								1: newLight=gt_replaceCoronaLightNodeVRayLight node	-- Uniform Spherical
								2: newLight=gt_replaceCoronaSunLightNodeVRayLight node
							)
						)
					)

					if (newLight!=undefined) do (
						if debug do
							print (node.name)
					)

					newLight!=undefined
				)
			),	
		
		fn gt_lightsToVRay selectedOnly = 
			(
				if debug do
					format " ********** Lights Conversion ********** \n\n"
				
				local oldLights=#()

				-- First traverse scene for daylight systems otherwise the sun/sky lights in them will appear as standalone
				-- lights to the normal lights replace traversal.
				for i=1 to $helpers.count do 
					(
						if (classof $helpers[i]==DaylightAssemblyHead) do (
							gt_replaceLight_mr_Sun $helpers[i]
						)
					)

					if ((environmentMap!=undefined) and (classof environmentMap==mr_Physical_Sky)) do (
						environmentMap=VRaySky()
					)
				

				if debug do
					if selectedOnly do  format "Converting scene Lights for selected lights...\n"
						
				for i=1 to $lights.count do 
					(
						-- update progress info
						gui_CoronaToVrayConverter.lbl_info.text = "Converting Light " + (i as string) + " of " + ($lights.count as string)
						gui_CoronaToVrayConverter.prog_convert.value = 100. * i/$lights.count	


					
					if selectedOnly then (
						if $lights[i].isSelected do
							if (gt_replaceLightNode $lights[i]) do ( append oldLights $lights[i] )
					) else (
						if (gt_replaceLightNode $lights[i]) do ( append oldLights $lights[i] )
					)
				)
				
				gui_CoronaToVrayConverter.lbl_info.text = ""
				gui_CoronaToVrayConverter.prog_convert.value = 0	
				
				-- old lights need to stay until the conversion is done, because deleteing them
				-- on the run messes nodes' order during traversal
		-- 			format "Done\n"
				if debug do
					format " ********** Finished Lights Conversion ********** \n\n"
			),				
			
			
		-- *********************************************************************** Bitmap Functions ***************************************************************************
			
		fn gt_convertB2HDRI_Texmap texmap=
			(
				undo off
				(
					if (texmap!=undefined and (classof texmap)==Bitmaptexture) then (
						local tex=undefined
						local mapFileNameIsValid=(texmap.fileName != undefined and texmap.fileName != "")
						--tex.gamma=1/2.2
						--tex.color_space=3 --Lele: set VRayBitmap's Color space to "From Max"

						--Here we would like to create VRayBitmap texture without Bitmaptexture file convertion.
						tex=VRayBitmap()
						tex.name=texmap.name
						if (mapFileNameIsValid) then (
							tex.HDRIMapName=texmap.fileName
						) else (
							format "Image path for % does not exist.\n" texmap
						)

						tex.color_space = 1 -- Inverse gamma; set this after setting HDRIMapName so as to override any automatic gamma detection in VRayBitmap
						local gamma=vrayGetBitmapGamma texmap
						if (gamma!=undefined) then (
							tex.gamma=1.0/gamma
						)

						if (tex!=undefined) then (
							tex.mapType=4
							tex.UVGen.blur=texmap.coords.blur
							tex.UVGen.mapping=texmap.coords.mapping
							tex.UVGen.mapChannel=texmap.coords.mapChannel
							tex.UVGen.mappingType=texmap.coords.mappingType
							tex.UVGen.UVW_Type=texmap.coords.UVW_Type
							tex.UVGen.U_Mirror=texmap.coords.U_Mirror
							tex.UVGen.V_Mirror=texmap.coords.V_Mirror
							tex.UVGen.U_Tile=texmap.coords.U_Tile
							tex.UVGen.V_Tile=texmap.coords.V_Tile
							tex.UVGen.showMapOnBack=texmap.coords.showMapOnBack
							tex.UVGen.Noise_On=texmap.coords.Noise_On
							tex.UVGen.Noise_Animate=texmap.coords.Noise_Animate
							tex.UVGen.realWorldScale=texmap.coords.realWorldScale
							tex.UVGen.realWorldHeight=texmap.coords.realWorldHeight
							tex.UVGen.realWorldWidth=texmap.coords.realWorldWidth
							tex.UVGen.phase=texmap.coords.phase
							tex.UVGen.U_Angle=texmap.coords.U_Angle
							tex.UVGen.V_Angle=texmap.coords.V_Angle
							tex.UVGen.W_Angle=texmap.coords.W_Angle
							tex.UVGen.Noise_Amount=texmap.coords.Noise_Amount
							tex.UVGen.Noise_Size=texmap.coords.Noise_Size
							tex.UVGen.Noise_Levels=texmap.coords.Noise_Levels
							tex.UVGen.Blur_Offset=texmap.coords.Blur_Offset
							tex.UVGen.u_tiling=texmap.coords.u_tiling
							tex.UVGen.v_tiling=texmap.coords.v_tiling
							tex.UVGen.u_offset=texmap.coords.u_offset
							tex.UVGen.v_offset=texmap.coords.v_offset

							--Convert crop/placement
							tex.cropplace_on = texmap.apply
							tex.cropplace_mode = texmap.cropPlace
							tex.cropplace_u = texmap.clipu
							tex.cropplace_v = texmap.clipv
							tex.cropplace_width = texmap.clipw
							tex.cropplace_height = texmap.cliph

							--Convert channel output
							tex.rgbOutput = texmap.rgbOutput
							tex.monoOutput = texmap.monoOutput
							tex.alphaSource = texmap.alphaSource

							--Convert output parameters including curve control
							replaceInstances tex.output (copy texmap.output)

							--Frees up all the memory used by the image file bitmap caches.
							--This is useful if memory is fragmented with a lot of different bitmaps and you want to have just the ones currently active reloaded.
							freescenebitmaps()
							--invokes the garbage collector without flush the undo system
							gc lite:true
						)
						tex
					)
					else (
						undefined
					)
				)
			),

		fn gt_convertB2HDRI_SubTex texmap = 
			(
				if (texmap!=undefined) then (
					-- check if we have already converted this texture; if yes, just return
					-- the already converted one
					local i=findItem glsc_orig_bitmaptextures texmap
					if i>0 then return glsc_new_hdritexmaps[i]

					for i = 1 to (getNumSubTexmaps texmap) do (
						local tex=(gt_convertB2HDRI_SubTex (getSubTexmap texmap i))
						if (tex!=undefined) do (
							setSubTexmap texmap i tex
						)
					)

					local new_tex = texmap
					if ((classof texmap)==Bitmaptexture) then (
						new_tex=gt_convertB2HDRI_Texmap texmap
					)

					-- if converted, add to the list of converted textures
					if new_tex!=texmap then (
						append glsc_orig_bitmaptextures texmap
						append glsc_new_hdritexmaps new_tex
						return new_tex
					)
				)
				return undefined
			),

		fn gt_convertB2HDRI_SubMtl material = 
			(
				if (material!=undefined) do (
					for i=1 to (getNumSubMtls material) do (
						gt_convertB2HDRI_SubMtl (getSubMtl material i)
					)

					for i=1 to (getNumSubTexmaps material) do (
						local tex=(gt_convertB2HDRI_SubTex (getSubTexmap material i))
						if (tex!=undefined) do (setSubTexmap material i tex)
					)
				)
			),

		fn gt_convertB2HDRI_ArrayParameter propArray = 
			(
				if ((classof propArray)==ArrayParameter) then (
					for i=1 to propArray.count do (
						if (propArray[i]!=undefined) do (
							if ((superclassof propArray[i])==textureMap) then (
								hdriTex=gt_convertB2HDRI_SubTex propArray[i]
								if (hdriTex!=undefined) do propArray[i]=hdriTex
							) else if((superclassof propArray[i])==material) then (
								gt_convertB2HDRI_SubMtl propArray[i]
							) else gt_convertB2HDRI_ArrayParameter propArray[i]
						)
					)
				)
			),

		fn gt_convertB2HDRI_Objects obj = 
			(
				undo off
				(
					if(obj != undefined) do (
						local specificObj
						local classOfObj = classof obj
						if ((superclassof obj)==light) then (
							if (classOfObj == miAreaLightOmni) then (
								specificObj = obj.Omni_Light
							)
							else if (classOfObj == miAreaLight) then (
								specificObj = obj.Target_Spot
							)
							else (
								specificObj = obj
							)
						)
						else (
							specificObj = obj
						)

						local listPropNames = getPropNames specificObj

						local hdriTex = undefined
						local propValue = undefined

						for propName in listPropNames do (
							if (isProperty specificObj propName) then (
								propValue = getProperty specificObj propName
								if ((superclassof propValue)==textureMap) then (
									hdriTex = gt_convertB2HDRI_SubTex propValue
									if (hdriTex!=undefined) do setProperty specificObj propName hdriTex
								) else if((superclassof propValue)==material) then (
									gt_convertB2HDRI_SubMtl propValue
								) else gt_convertB2HDRI_ArrayParameter propValue
							)
						)
					)
				)
			),

		fn gt_convertB2HDRI_Node node applyOnObjects = 
			(
				if isValidNode node then (
					gt_convertB2HDRI_SubMtl node.material
					if (applyOnObjects) do (gt_convertB2HDRI_Objects node.baseObject)
					--convert bitmaps of node modifiers
					for imod in node.modifiers do (
						gt_convertB2HDRI_Objects imod
					)
				)
			),

		fn gt_convertB2HDRI_RenderSettings = 
			(
				undo off
				(
					if(vrayCheckProductionRenderer()) then (
						productRenderer = renderers.production
						listRenderSettingsNames = getPropNames productRenderer
						for rParamName in listRenderSettingsNames do (
							if (isProperty productRenderer rParamName) then (
								rParam = getProperty productRenderer rParamName
								if ((superclassof rParam)==textureMap) then (
									hdriTex = gt_convertB2HDRI_SubTex rParam
									if (hdriTex!=undefined) do setProperty productRenderer rParamName hdriTex
								) else if ((superclassof rParam)==material) then (
									gt_convertB2HDRI_SubMtl rParam
								) else gt_convertB2HDRI_ArrayParameter rParam
							)
						)
					)
				)
			),

		fn gt_convertB2HDRI_Environment = 
			(
				undo off
				(
					-- Convert Environment map
					local envtex=gt_convertB2HDRI_SubTex environmentMap
					if (envtex!=undefined) do environmentMap=envtex

					-- Convert Atmosphere effects
					for iAtmos=1 to numAtmospherics do (
						currAtmospheric = getAtmospheric iAtmos
						listAtmosNames = getPropNames currAtmospheric

						for atmosName in listAtmosNames do (
							if (isProperty currAtmospheric atmosName) then (
								atmosParam = getProperty currAtmospheric atmosName
								if ((superclassof atmosParam)==textureMap) then (
									hdriTex = gt_convertB2HDRI_SubTex atmosParam
									if (hdriTex!=undefined) do setProperty currAtmospheric atmosName hdriTex
								) else if ((superclassof atmosParam)==material) then (
									gt_convertB2HDRI_SubMtl atmosParam
								) else gt_convertB2HDRI_ArrayParameter atmosParam
							)
						)
					)

					-- Convert Environment effects
					for iEffect=1 to numEffects do (
						currEffect = getEffect iEffect
						listEffectNames = getPropNames currEffect
						for effectName in listEffectNames do (
							if (isProperty currEffect effectName) then (
								effectParam = getProperty currEffect effectName
								if ((superclassof effectParam)==textureMap) then (
									hdriTex = gt_convertB2HDRI_SubTex effectParam
									if (hdriTex!=undefined) do setProperty currEffect effectName hdriTex
								) else if ((superclassof effectParam)==material) then (
									gt_convertB2HDRI_SubMtl effectParam
								) else gt_convertB2HDRI_ArrayParameter effectParam
							)
						)
					)
				)
			),

		--The main function that visits scene objects and replace Bitmaptexture with corresponding VRayBitmap
		--exrConvert:<true|false> - true->converts bitmaps files to EXR/TX files.
		--applyOnSceneObjects:<true|false> -> true->allows Bitmaptextures directly assigned to scene nodes to be converted to VRayBitmap
		fn gt_VRayBitmap2VRayBitmap =
			(
				local listBmpTexs = getClassInstances bitmaptex processChildren:true

				with redraw off (
					for iTex in listBmpTexs do (
						new_tex=gt_convertB2HDRI_Texmap iTex
						if new_tex!=undefined then (
							replaceinstances iTex new_tex
						)
						else format "Converting Bitmap % to VRayBitmap was unsuccessful\n" iTex
					)
				)

				--here we must avoid Bitmaptextures directly assigned to scene nodes
				glsc_new_hdritexmaps=#()
				glsc_orig_bitmaptextures=#()
				glsc_listOfConvertedNodesHandles=#()

				-- Disable scene redraw during the replacement in order to prevent 3ds Max from updating the viewports after every change
				with redraw off (
					for iTex in listBmpTexs do (
						local listOfNodes =  (refs.dependentNodes iTex)
						for iNode in listOfNodes do (
							local i=findItem glsc_listOfConvertedNodesHandles iNode.handle
							if i == 0 do (
								gt_convertB2HDRI_Node iNode applyOnObjects
								append glsc_listOfConvertedNodesHandles iNode.handle
							)
						)
					)

					for i=1 to meditMaterials.count do (
						local m=meditMaterials[i]
						if (m.superClassID==3072) then ( -- This is a material
							gt_convertB2HDRI_SubMtl m
						) else ( -- This is a texture
							m=gt_convertB2HDRI_SubTex m
							if (m!=undefined) then meditMaterials[i]=m
						)
						updateMTLInMedit meditMaterials[i]
					)

					gt_convertB2HDRI_RenderSettings()
					gt_convertB2HDRI_Environment()
				)
			),		

		-- *********************************************************************** Slate Material Editor Functions ***************************************************************
		-- check if there are maps in the enviorments.  vray must be the current renderd
		fn isEnviormentMap theRef =
		(
			result = false
			
			if theRef == environmentMap then 
				result = true
			
			if renderers.current.environment_gi_map != undefined then (
				if theRef == renderers.current.environment_gi_map then 
					result = true
			)
			
			if renderers.current.environment_rr_map != undefined then (
				if theRef == renderers.current.environment_rr_map then 
					result = true
			)

			if renderers.current.environment_refract_map != undefined then (
				if theRef == renderers.current.environment_refract_map then 
					result = true
			)

			if renderers.current.environment_secondaryMatte_map != undefined then (
				if theRef == renderers.current.environment_secondaryMatte_map then
					result = true
			)
			
			return result
		),
		
		-- function to recursively find submaps to remove in an array
		fn gt_isMapInArray &theArray texmap debug:false = 
			(
				if (texmap!=undefined) then
				(
					id = finditem theArray texmap
					if id != 0 then 
						(
							if debug do
								format "    Exluding Node: %\n" texmap
							deleteitem theArray id
							
							-- check the map type
							for i = 1 to (getNumSubTexmaps texmap) do (
								theSubTex = getSubTexmap texmap i
								if theSubTex != undefined then
-- 									format "   here: % \n" theSubTex
									gt_isMapInArray theArray theSubTex
							)
						)
				)
			),
			
		-- function to cleanup the slate editor by removeing unsed maps.
		-- I need to do further testing on this one
		fn removeUnsuedNodesfromSlateEditor debug:false =
		(
			if debug do
				format " ********** Cleaning Up Slate Editor ********** \n\n"

			smViewCount = sme.GetNumViews()

			for i=1 to sme.GetNumViews() do 
			(					
				smView = sme.GetView(i) -- sme.GetView i
				smView.SelectNone()

				if debug do
					format "  Tab : % , nodes: % \n" smView.name (smView.GetNumNodes() as string)
				
				local nodesToDelete = #()

				
				for n = 1 to smView.GetNumNodes() do
					(
						-- this node has these proerties #(#name, #position, #selected, #reference)
						currentNode = (smView.GetNode  n)
						theRef = currentNode.reference

						nodes = refs.dependentnodes theRef
						-- add node to delete array
						if nodes.count == 0 then (
							if debug do
								format "   Collecting Node:  % \n" theRef
								append nodesToDelete theRef		
								)
					)
					
					
					for n = 1 to smView.GetNumNodes() do
						(
							-- this node has these proerties #(#name, #position, #selected, #reference)
							currentNode = (smView.GetNode  n)
							theRef = currentNode.reference

							if ((isEnviormentMap theRef) == true) then 
								(
									gt_isMapInArray nodesToDelete theRef debug:debug
								)		
						)			

				smView.SelectNone()
				for o = 1 to nodesToDelete.count do
					(
						theNode = smView.GetNodeByRef nodesToDelete[o]
						theNode.selected = true
						format "     Deleting SME Node: % \n" theNode.name
					)						

-- 				format " nodesToDelete: %\n" nodesToDelete
				smView.DeleteSelection()
			)
				
			if debug do
				format " ********** End Cleaning Up Slate Editor ********** \n\n"			
		),
		
		fn gt_deleteSMESelectedNodesInEachView = 
			(
				local numViews = sme.GetNumViews()
				for iView=1 to numViews do (
					local currentView = sme.GetView iView
					currentView.DeleteSelection()
				)
			),		

		fn gt_selectSMENodeByRef objRef = 
			(
				local numViews = sme.GetNumViews()
				for iView=1 to numViews do (
					local currentView = sme.GetView iView
					local nodeByRef = currentView.GetNodeByRef objRef
					if (nodeByRef != undefined) then (
						nodeByRef.selected = true
						
						if debug do
							nodeByRef.name
					)
				)
			),
	
		--create a node in Slate Material Editor for
		--newly creted material/texture on the position of the old one
		--the node for objRef is not deleted
		fn gt_createSMENodeByRef objRef newObjRef  = 
			(
				if(objRef != newObjRef) do 
				(
					local numViews = sme.GetNumViews()
					
					for iView=1 to numViews do 
						(
							local currentView = sme.GetView iView
							local nodeByRef = currentView.GetNodeByRef objRef
							
							if (nodeByRef != undefined) then (
-- 								currentView.CreateNode newObjRef (nodeByRef.position - [200, -80]) -- move it a bit to the left
								currentView.CreateNode newObjRef nodeByRef.position 
								break
							)
						)
				)
			),
			
		fn gt_deselectSMENodes = 
			(
				local numViews = sme.GetNumViews()
				for iView=1 to numViews do (
					local currentView = sme.GetView iView
					local numNodes = currentView.GetNumNodes()
					for iNode=1 to numNodes do (
						local currentNode = currentView.GetNode iNode
						currentNode.selected = false
					)
				)
			),

		fn gt_showMtlSlotInSME activeSlot= 
			(
				local m = getMeditMaterial activeSlot
				if (MatEditor.mode == #advanced) then (
					local vrmatViewIdx = sme.GetViewByName "VRayMtlView"
					if (vrmatViewIdx == 0) then (
						vrmatViewIdx = sme.CreateView "VRayMtlView"
					)
					local vrmatView = sme.GetView vrmatViewIdx

					if (vrmatView != undefined) then (
						vrmatView.CreateNode m [0,0]
					)
				)
			),	

		fn gt_showMtlSlotInCurrentSMEView activeSlot= 
			(
				local m = getMeditMaterial activeSlot
				local vrmatViewIdx = sme.activeView
				if (vrmatViewIdx == 0) then (
					vrmatViewIdx = 1
				)

				local vrmatView = sme.GetView vrmatViewIdx
				if (vrmatView != undefined) then (
					vrmatView.CreateNode m [0,0]
				)
			),

		fn gt_mtlExistsInSME mtlName= 
			(
				for i=1 to sme.GetNumViews() do (
					view=sme.GetView(i)
					for j=1 to view.GetNumNodes() do (
						node=view.GetNode(j)
						if (node.name == mtlName) then (
							return true
						)
					)
				)
				return false
			),			
			
		-- *********************************************************************** Textures Functions **************************************************************************
			
		-- Function used in conversion from CoronaColorCorrect to Color_Correction
		-- when 'controlsSource' is set to 1 ->'Use settings from map: '. That means
		-- settings from another <set to 'copiedMap' parameter> CoronaColorCorrect
		-- map will be used. The function helps to reach the base map which settings are used.
		fn gt_getTexSettingsBase inTex =
			(
				if ((inTex != undefined) and (hasProperty inTex "controlsSource") and (hasProperty inTex "copiedMap")) do (
					local texSettings = inTex
					if inTex.controlsSource == 1 do ( -- 'Use settings from map'
						texSettings = gt_getTexSettingsBase inTex.copiedMap
					)
					return texSettings
				)
				undefined
			),
		
		fn gt_convertFrom_UnsupportedMap texmap =
			(
				if debug then
					format "   ****** HEY UNSUPORTED MAP : % \n" texmap
				
				local tmpColor = VRayColor()
				tmpColor.name = "Unsupported_Map_"+ texmap.name
				tmpColor.color = unsupportedColor 

				return texmap
			),	
			
		fn gt_convertFrom_CoronaBumpConverter inTexture = 
			(
				local vrColor2BumpTex = VRayColor2Bump name:inTexture.name
				vrColor2BumpTex.map = inTexture.input
						
				return vrColor2BumpTex
			),

		fn gt_convertFrom_CoronaColorTex inTexture  = 
			(

					vrayC = VRayColor name:inTexture.name alpha:inTexture.alpha rgb_multiplier: inTexture.multiplier
					if inTexture.method == 0 do (
						vrayC.color_mode = 0
						vrayC.color = inTexture.color
					)
					if inTexture.method == 1 do (
						vrayC.color_mode = 0
						vrayC.red = inTexture.colorHdr[1]
						vrayC.green = inTexture.colorHdr[2]
						vrayC.blue = inTexture.colorHdr[3]
					)
					if inTexture.method == 2 do (
						vrayC.color_mode = 1
						vrayC.temperature = inTexture.temperature
					)
					if inTexture.method == 3 do (
						vrayC.color_mode = 0
						local strHexColor = inTexture.hexColor
						if (strHexColor.count == 6 or strHexColor.count == 7) then (
							local startVal = 1
							if ((findString strHexColor "#") == 1) do startVal = 2

							local subStrHexColor = "0x" + (substring strHexColor startVal 2)
							local rColor = (subStrHexColor as float)
							startVal = startVal +2
							subStrHexColor = "0x" + (substring strHexColor startVal 2)
							local gColor = (subStrHexColor as float)
							startVal = startVal +2
							subStrHexColor = "0x" + (substring strHexColor startVal 2)
							local bColor = (subStrHexColor as float)

							if (rColor!=undefined and gColor!=undefined and bColor!=undefined) then (
								vrayC.color = (color rColor gColor bColor)
							) else (
								vrayC.color = (color 0.0f 0.0f 0.0f)
							)
						) else (
							vrayC.color = (color 255.0f 0.0f 0.0f)
						)
					)
							
					return  vrayC

			),
    
		-- TODO
		-- probably need to work more on this :)
		fn gt_convertFrom_CoronaFrontBack inTexture =
			(
				if inTexture.frontTexmap != undefined then
					return inTexture.frontTexmap
				else 
					gt_convertFrom_UnsupportedMap inTexture
			),
		
		fn gt_convertFrom_CoronaMultiMapTex inTexture  = 
			(
				vrayMST = VRayMultiSubTex name:inTexture.name
				if vrayMST.texmap_num > inTexture.items do (
					local removeItems = vrayMST.texmap_num - inTexture.items
					vrayMST.deleteSubtex (inTexture.items + 1) removeItems
				)

				if vrayMST.texmap_num < inTexture.items do (
					local addItems = inTexture.items - vrayMST.texmap_num
					vrayMST.addSubtex addItems
				)
				
				vrayMST.random_gamma = inTexture.gammaRandom
				vrayMST.random_saturation = inTexture.saturationRandom
				vrayMST.random_hue = inTexture.hueRandom
				
				/*
					CORONA
				randomizePrimitive			-- Primitive
				randomizeSubPrimitive		-- Sub Primitive
				randomizePolygon				-- Polygon
				randomizeMeshElement		-- Mesh Element
				randomizeDecal				-- Decal
					
				randomizeInstance			-- Instance
				randomizeMaterialGBuffer	-- Material GBuffer ID
				randomizeObjectGBuffer	-- Object GBuffer ID
				randomizeMaterialName		-- Material Name
				randomizeMaterialId 		-- Face Material ID

				VRAY			
				from_id = 0									-- Face Material ID
				from_id = 1									-- Object ID

				from_id = 12									-- Random
				random_by_name = on					-- By Name	
				random_by_instance_id					-- By Instance ID
				random_by_object_id					-- By Object ID
				random_by_element						-- By Elements
				random_by_particle_id					-- By Particle ID
				random_by_node_handle					-- By Node Handle
				random_by_face_mtl_id					-- By Face ID
				random_by_render_id					-- By Render ID
				random_by_user_id						-- By User ID
				*/
				-- 			if inTexture.mode == 5 do vrayMST.from_id = 1
				-- 			if inTexture.mode == 2 do vrayMST.from_id = 6


				vrayMST.from_id = 12 -- Random
				
				if inTexture.randomizeMeshElement 	== on do vrayMST.random_by_element = on		-- Mesh Element
				if inTexture.randomizeMaterialId 			== on do vrayMST.random_by_face_mtl_id = on	-- Face Material ID
				if inTexture.randomizeInstance			== on do vrayMST.random_by_instance_id = on	-- Instance
				if inTexture.randomizeObjectGBuffer	== on do vrayMST.random_by_object_id = on		-- Object GBuffer ID
				if inTexture.randomizeMaterialName		== on do vrayMST.random_by_name = on			-- Material Name

				for i = 1 to inTexture.items do (
					vrayMST.texmap_color[i] = inTexture.colors[i]
					vrayMST.texmap[i] = inTexture.texmaps[i]
				)
						
				return vrayMST
			),
    
		fn gt_convertFrom_CoronaMixTex inTexture = 
			(
				vrayCT = VRayCompTex sourceA: inTexture.texmapBottom sourceB: inTexture.texmapTop name:inTexture.name
				vrayCT.operator = case inTexture.mixOperation of (
					0: 0 -- Add
					1: 1 -- subtract
					2: 3 -- multiply
					3: 4 -- divide
					4: 5 -- minimum
					5: 6 -- maximum
					8: 2 -- difference
					default:0
					)
					
				return vrayCT
			),
    
		fn gt_convertFrom_CoronaAoTex inTexture =
			(
				vrayDt = VRayDirt occluded_color:inTexture.colorOccluded unoccluded_color:inTexture.colorUnoccluded \
				radius:inTexture.maxDistance texmap_radius:inTexture.texmapDistance texmap_occluded_color:inTexture.texmapOccluded \
				texmap_unoccluded_color:inTexture.texmapUnoccluded name:inTexture.name

				if (hasProperty inTexture "invert") then (
					vrayDt.invert_normal = inTexture.invert
				)
				else if (hasProperty inTexture "normalMode") then (
					vrayDt.invert_normal = case of (
						(inTexture.normalMode == 0):false
						(inTexture.normalMode == 1):true
						(inTexture.normalMode == 2):false
						default:false)
				)

				if isCorona10orHiger() then (
					if inTexture.occlusionFrom == 1 and inTexture.excludeList.count != 0 do vrayDt.excludeList = inTexture.excludeList
					if inTexture.occlusionFrom == 3 do vrayDt.consider_same_object_only = true

				) else (
					if inTexture.excludeMode == 1 and inTexture.excludeList.count != 0 do vrayDt.excludeList = inTexture.excludeList
					if inTexture.excludeMode == 3 do vrayDt.consider_same_object_only = true
				)

				vrayDt.texmap_radius = if inTexture.texmapDistance != undefined do inTexture.texmapDistance
				vrayDt.texmap_occluded_color = if inTexture.texmapOccluded != undefined do inTexture.texmapOccluded
				vrayDt.texmap_unoccluded_color = if inTexture.texmapUnoccluded != undefined do inTexture.texmapUnoccluded
					
				return vrayDt
			),
    
		fn gt_convertFrom_CoronaNormalTex inTexture = 
			(
				vrayNM = VRayNormalMap normal_map:inTexture.normalMap name:inTexture.name

				if ((hasProperty vrayNM #flip_red) and
					(hasProperty inTexture #flipred)) then (
						vrayNM.flip_red = inTexture.flipred
						vrayNM.flip_green = inTexture.flipgreen
						vrayNM.swap_red_and_green = inTexture.swapRedGreen
				)

				if ((hasProperty vrayNM #bump_map) and -- Vray 3.3
					(hasProperty inTexture #additionalBump)) then (
					vrayNM.bump_map = inTexture.additionalBump
					if vrayNM.bump_map != undefined do (
						vrayNM.bump_map_on = inTexture.additionalBumpOn
						vrayNM.bump_map_multiplier = inTexture.additionalBumpStrength
					)
				)

				if (hasProperty vrayNM #apply_gamma) then (
					vrayNM.apply_gamma = inTexture.addGamma
				)
					
				return vrayNM
			),

		fn gt_convertFrom_CoronaSky inTexture =
			(
				local tmpSky = VRaySky()
				tmpSky.name = inTexture.name
					
				return tmpSky
			),
		
		-- TODO:
		-- I need to work more on this
		fn gt_convertFrom_CoronaCurvatureTex inTexture =
			(
				local tmpTex = VRayCurvature()
				tmpTex.name =  inTexture.name
				
				tmpTex.radius = inTexture.distanceMultiplier
					
				return tmpTex	
			),
			
		fn gt_convertFrom_CoronaWireTex inTexture =
			(
				local tmpVRayEdgesTex =  VRayEdgesTex ()
				tmpVRayEdgesTex.name = inTexture.name
				tmpVRayEdgesTex.edgesColor = inTexture.edgeColor 
				tmpVRayEdgesTex.widthType = (if inTexture.widthMode == 1 then 0 else 1) 
				tmpVRayEdgesTex.thickness = inTexture.edgeWidthWorld 
				tmpVRayEdgesTex.pixelWidth = inTexture.edgeWidthPixels 
					
				return tmpVRayEdgesTex
			),
		
		fn gt_convertFrom_CoronaRoundEdges inTexture  =
			(
			vrRoundEdges = VRayEdgesTex name:inTexture.name roundedCorners_radius:inTexture.radius roundedCorners_sameObjectOnly:(inTexture.excludeMode == 1)
			
			return vrRoundEdges
					
			),
    
		fn gt_convertFrom_CoronaDistance inTexture = 
			(
				vrDistTex = VRayDistanceTex  name:inTexture.name

				vrDistTex.objects = inTexture.nodes
				vrDistTex.texmap_far_on = inTexture.texmapFarOn
				vrDistTex.far_color = inTexture.colorFar
				vrDistTex.texmap_far = inTexture.texmapFar
				vrDistTex.texmap_near_on = inTexture.texmapNearOn
				vrDistTex.near_color = inTexture.colorNear
				vrDistTex.texmap_near = inTexture.texmapNear
				vrDistTex.distance = inTexture.distanceFar
				vrDistTex.tex_inside_separate_on = inTexture.useInside
				vrDistTex.texmap_distance_on = inTexture.distanceScaleTexmapOn
				vrDistTex.texmap_distance = inTexture.distanceScaleTexmap
				
				vrDistTex.inside solid = on

				if (inTexture.useInside == on) then (
					vrDistTex.inside_color = inTexture.colorInside
					vrDistTex.texmap_inside = inTexture.texmapInside
					vrDistTex.texmap_inside_on = inTexture.texmapInsideOn
				)
					
				vrDistTex
			),	
		
		fn gt_convertFrom_CoronaTriplanar inTexture = 
			(
				vrTriplanarTex = VRayTriplanarTex name:inTexture.name

				vrTriplanarTex.texture_mode = if inTexture.onlyTexmapX == on then 0 else 1

				if inTexture.texmapX !=undefined and inTexture.texmapXOn == on then
					vrTriplanarTex.texture = inTexture.texmapX
				if inTexture.texmapY !=undefined and inTexture.texmapYOn == on then
					vrTriplanarTex.texture_y = inTexture.texmapY
				if inTexture.texmapZ !=undefined and inTexture.texmapZOn == on then
					vrTriplanarTex.texture_z = inTexture.texmapZ

				vrTriplanarTex.size = inTexture.scale
				vrTriplanarTex.blend = inTexture.blend

				case inTexture.space of (
					0: (
						vrTriplanarTex.space = 0
					)
					1: (
						vrTriplanarTex.space = 1
						vrTriplanarTex.reference_node = undefined
					)
					2: (
						vrTriplanarTex.space = 1
						vrTriplanarTex.reference_node = inTexture.referenceNode
					)
				)
				vrTriplanarTex.frame_offset = inTexture.offset
				vrTriplanarTex.texture_rotation = inTexture.rotation
					
				vrTriplanarTex
			),

		fn gt_convertFrom_CoronaBitmap inTexture = 
			(
				local texHDRI=VRayBitmap()
				texHDRI.name=inTexture.name

				if inTexture.filename != "" and inTexture.filename != undefined then (
					texHDRI.HDRIMapName = inTexture.filename
				)

				--Corona Bitmap -> Environment mode 4: MirrorBall,  corresponds to
				--VRay Bitmap -> Mapping type 3: Mirrored ball
				if inTexture.enviroMapping == 4 then texHDRI.mapType = 3

				--Corona Bitmap -> Environment mode 0: Spherical
				--VRay Bitmap -> Mapping type 4: 3ds Max standard
				if inTexture.enviroMapping == 0 then (
					texHDRI.mapType = 4 --3ds Max standard
					if environmentMap == inTexture then
						texHDRI.coords.mappingType = 1 -- Coordinates -> Environ
					else
						texHDRI.coords.mappingType = 0 -- Coordinates -> Texture
					texHDRI.coords.mapping = 0 -- Coordinates -> Spherical Environment
				)

				case inTexture.interpolation of (
					0: (texHDRI.interpType = 0; texHDRI.filterMode = 0)
					1: (texHDRI.interpType = 1; texHDRI.filterMode = 0)
					2: texHDRI.filterMode = 1
				)

				-- Gamma handling
				-- Set this after setting HDRIMapName so as to override any automatic gamma detection in VRayBitmap
				if (maxversion())[1] >= 16000 and inTexture.gamma != -1 then(
					texHDRI.color_space = 1 -- Inverse gamma
					texHDRI.gamma = 1.0/inTexture.gamma
				) else (
					texHDRI.color_space = 3 -- Automatic gamma
				)

				-- uv channel, environment mode
				if texHDRI.coords.mappingType==0 and inTexture.uvwChannel == 0 then (
					texHDRI.coords.mapping = 1
				)else(
					texHDRI.coords.mapChannel = inTexture.uvwChannel
				)

				-- offset, tiling/scale, rotation, rw mapscale
				texHDRI.coords.realWorldScale = inTexture.realWorldScale

				texHDRI.coords.U_Offset = inTexture.uvwOffset[1]
				texHDRI.coords.V_Offset = inTexture.uvwOffset[2]

				if inTexture.realWorldScale then (
					texHDRI.coords.realWorldWidth = inTexture.uvwScale[1]
					texHDRI.coords.realWorldHeight = inTexture.uvwScale[2]
				)else(
					texHDRI.coords.U_Tiling = inTexture.uvwScale[1]
					texHDRI.coords.V_Tiling = inTexture.uvwScale[2]
				)

				case inTexture.tilingU of (
					0: (texHDRI.coords.U_Mirror = false; texHDRI.coords.U_Tile = false)
					2: (texHDRI.coords.U_Mirror = true; texHDRI.coords.U_Tile = false)
					1: (texHDRI.coords.U_Mirror = false; texHDRI.coords.U_Tile = true)
				)
				case inTexture.tilingV of (
					0: (texHDRI.coords.V_Mirror = false; texHDRI.coords.V_Tile = false)
					2: (texHDRI.coords.V_Mirror = true; texHDRI.coords.V_Tile = false)
					1: (texHDRI.coords.V_Mirror = false; texHDRI.coords.V_Tile = true)
				)
				texHDRI.coords.W_angle = inTexture.wAngle
				if environmentMap == inTexture then (
					texHDRI.coords.W_angle *= -1.0f
				)

				-- crop/place
				texHDRI.cropplace_on = inTexture.clippingOn
				texHDRI.cropplace_mode = inTexture.clippingMode
				texHDRI.cropplace_u = inTexture.clippingU
				texHDRI.cropplace_v = inTexture.clippingV
				texHDRI.cropplace_width = inTexture.clippingWidth
				texHDRI.cropplace_height = inTexture.clippingHeight

				-- alpha source, mono channel out, rgb channel out
				texHDRI.alphaSource = inTexture.alphaSource
				texHDRI.monoOutput = inTexture.monoChannelOutput
				texHDRI.rgbOutput = inTexture.rgbChannelOutput

				texHDRI.coords.blur = inTexture.filteringBlur

				--Convert output parameters including curve control
				replaceInstances texHDRI.output (copy inTexture.output)
					
				texHDRI
			),

		fn gt_convertFrom_CoronaColorCorrect inTexture = 
			(
				if gui_CoronaToVrayConverter.chk_CrnColorCorrect.checked then
					(
						local colorCorrectionTex = Color_Correction name:inTexture.name
						colorCorrectionTex.map = inTexture.inputTexmap

						-- Get base map which settings will be used
						local texSettings = gt_getTexSettingsBase inTexture
						format "texSettings:  % \n" texSettings
						
						if texSettings != undefined then ( -- 'Use these settings:' option in CoronaColorCorrect
							colorCorrectionTex.hueShift = texSettings.hue
							colorCorrectionTex.saturation = texSettings.saturation
							
							colorCorrectionTex.lightnessMode=1 -- Lightness advanced mode
							-- Additional correctons for brightness and contrast
		-- 					colorCorrectionTex.liftRGB = texSettings.brightness
							colorCorrectionTex.gammaRGB = (texSettings.gamma + texSettings.contrast)/2 - .05f --- 0.5f
							
							if colorCorrectionTex.gammaRGB >1 do (
								colorCorrectionTex.gammaRGB = colorCorrectionTex.gammaRGB/10.0f
								colorCorrectionTex.gammaRGB = colorCorrectionTex.gammaRGB  + 0.5f
							)

							if texSettings.invert == on do colorCorrectionTex.rewireMode = 2 -- Channels->Invert
							if texSettings.lutEnable == true and texSettings.lutFile != undefined then (
								local vrayLUTTex = VRayLUT()
								colorCorrectionTex.gammaRGB = colorCorrectionTex.gammaRGB - 0.5f
								vrayLUTTex.use_logspace = texSettings.lutLog
								if vrayLUTTex.use_logspace == on then (
									colorCorrectionTex.gammaRGB = colorCorrectionTex.gammaRGB +1.0f
								)
								vrayLUTTex.basemap = colorCorrectionTex
								vrayLUTTex.lutFile = texSettings.lutFile
								return vrayLUTTex
							)
						)
							
						colorCorrectionTex
					)else
							return inTexture
			),

		fn gt_convertFrom_NormalBump inTexture  =
			(
				local texNormalBump = VRayNormalMap()

				texNormalBump.name = inTexture.name 
				texNormalBump.normal_map = inTexture.normal_map 
				texNormalBump.normal_map_on = inTexture.map1on 
				texNormalBump.normal_map_multiplier = inTexture.mult_spin 
				texNormalBump.bump_map = inTexture.bump_map 
				texNormalBump.bump_map_on = inTexture.map2on
				texNormalBump.bump_map_multiplier = inTexture.bump_spin
				texNormalBump.flip_red = inTexture.flipred 
				texNormalBump.flip_green = inTexture.flipgreen 
				texNormalBump.swap_red_and_green = inTexture.swap_rg
						
				return texNormalBump
			),
		
		fn gt_convertFrom_AutodeskBitmap texmap =
			(
				if (texmap!=undefined and (classof texmap)==Autodesk_Map) then (
					local texHDRI=VRayBitmap()
					texHDRI.name=texmap.name

					if (IsProperty texmap "Parameters_Source") and (undefined != texmap.Parameters_Source) do
					(
						texHDRI.HDRIMapName = texmap.Parameters_Source.filename
					)
					if (IsProperty texmap "Parameters_Brightness") and (undefined != texmap.Parameters_Brightness) do
					(
						texHDRI.output.RGB_Level = texmap.Parameters_Brightness / 100
					)

					if (IsProperty texmap "Parameters_Invert_Image") and (undefined != texmap.Parameters_Invert_Image) do
					(
						texHDRI.output.invert = texmap.Parameters_Invert_Image
					)

					texHDRI.coords.mappingType = 0 --Texture
					texHDRI.coords.mapping = 0 --Explicit Map Channel
					texHDRI.coords.realWorldScale = true

					if (IsProperty texmap "Position_X") and (undefined != texmap.Position_X) do
					(
						texHDRI.UVGen.u_offset = texmap.Position_X
						if(texmap.Position_X.controller != undefined) do texHDRI.UVGen.u_offset = texmap.Position_X.controller
					)

					if (IsProperty texmap "Position_Y") and (undefined != texmap.Position_Y) do
					(
						texHDRI.UVGen.v_offset = texmap.Position_Y
						if(texmap.Position_Y.controller != undefined) do texHDRI.UVGen.v_offset.controller = texmap.Position_Y.controller
					)

					if (IsProperty texmap "Scale_Width") and (undefined != texmap.Scale_Width) do
					(
						texHDRI.UVGen.realWorldWidth = texmap.Scale_Width
						if(texmap.Scale_Width.controller != undefined) do texHDRI.UVGen.realWorldWidth.controller = texmap.Scale_Width.controller
					)

					if (IsProperty texmap "Scale_Height") and (undefined != texmap.Scale_Height) do
					(
						texHDRI.UVGen.realWorldHeight = texmap.Scale_Height
						if(texmap.Scale_Height.controller != undefined) do texHDRI.UVGen.realWorldHeight.controller = texmap.Scale_Height.controller
					)

					--According to the documentation the rotation is done on the Z axis, so rotation is on the W axis
					if (IsProperty texmap "Position_Rotation") and (undefined != texmap.Position_Rotation) do
					(
						texHDRI.UVGen.W_Angle = texmap.Position_Rotation
						if(texmap.Position_Rotation.controller != undefined) do texHDRI.UVGen.W_Angle.controller = texmap.Position_Rotation.controller
					)
					texHDRI.UVGen.V_Angle = 0.0
					texHDRI.UVGen.U_Angle = 0.0

					texHDRI.UVGen.U_Mirror = false
					texHDRI.UVGen.V_Mirror = false
					if (IsProperty texmap "Repeat_Horizontal") and (undefined != texmap.Repeat_Horizontal) do
					(
						texHDRI.UVGen.U_Tile = texmap.Repeat_Horizontal
					)
					if (IsProperty texmap "Repeat_Vertical") and (undefined != texmap.Repeat_Vertical) do
					(
						texHDRI.UVGen.V_Tile = texmap.Repeat_Vertical
					)

					if (IsProperty texmap "Advanced_Parameters_Blur") and (undefined != texmap.Advanced_Parameters_Blur) do
					(
						texHDRI.UVGen.blur = texmap.Advanced_Parameters_Blur
						if(texmap.Advanced_Parameters_Blur.controller != undefined) do texHDRI.UVGen.blur.controller = texmap.Advanced_Parameters_Blur.controller
					)

					if (IsProperty texmap "Advanced_Parameters_Blur_Offset") and (undefined != texmap.Advanced_Parameters_Blur_Offset) do
					(
						texHDRI.UVGen.Blur_Offset = texmap.Advanced_Parameters_Blur_Offset
						if(texmap.Advanced_Parameters_Blur_Offset.controller != undefined) do texHDRI.UVGen.Blur_Offset.controller = texmap.Advanced_Parameters_Blur_Offset.controller
					)

					if (IsProperty texmap "Advanced_Parameters_Map_Channel") and (undefined != texmap.Advanced_Parameters_Map_Channel) do
					(
						texHDRI.UVGen.mapChannel = texmap.Advanced_Parameters_Map_Channel
					)

					texHDRI.mapType=4 /*The mapping type is determined by the Coordinates section.*/
					texHDRI.color_space=3 /*The gamma is controlled from 3ds Max (customize -> preferences -> Gamma and LUT -> Input Gamma)*/

					
					texHDRI
				)
				else (
					undefined
				)
			),
		
		fn gt_convertFrom_RaytraceMap texmap =	
			(
				
			),
			
		fn gt_isVRayCompatible mtl_classid = 
			(
				-- native V-Ray materials
				if mtl_classid==VRayMtl then return true
				if mtl_classid==VRay2SidedMtl then return true
				if mtl_classid==VRayBlendMtl then return true
				if mtl_classid==VRayFastSSS then return true
				if mtl_classid==VRayLightMtl then return true
				if mtl_classid==VRayMtlWrapper then return true

				-- supported 3dsmax materials
				if mtl_classid==Multimaterial then return true
				if mtl_classid==Arch___Design__mi then return true

				false
			),	

		fn gt_isVRayCompatibleTex tex_classid = 
			(
				-- native V-Ray textures
				compatibleTexClassIDs = #(VRayBmpFilter, VRayBump2Normal, VRayColor, VRayColor2Bump, VRayCompTex, VRayCurvature, VRayDirt, VRayDistanceTex, VRayEdgesTex, \
														VRayFakeFresnelTex, VRayGLSLTex, VRayHairInfoTex, VRayBitmap, VRayICC, VRayLut, VRayMap, VRayMultiSubTex, VRayOCIO, VRayOSLTex, \
														VRayParticleTex, VRayPtex, VRaySamplerInfoTex, VRaySky, VRaySoftbox, VRayTriplanarTex, VRayUserColor, VRayUserScalar)
				result = findItem compatibleTexClassIDs tex_classid
				if result > 0 then return true
				return false
			),

		fn testingTex inTexture =
		(
			format " ***** HEY !! : %\n" (classof inTexture) 
			
			return inTexture
		),
		
		fn gt_createVRayTexture inTexture= 
			(
				undo off
				(
				-- check if this is a texture; if not, just return the value
				if ((superclassof inTexture)!=textureMap) then return inTexture

				-- check if we have already converted this texture; if yes, just return
				-- the already converted one
				local i=findItem glsc_orig_texmaps inTexture
				if i>0 then return glsc_new_texmaps[i]

				-- recurse into sub-textures, if any
				local sub_count=getNumSubTexmaps inTexture
				if sub_count>0 then (
					local sub_tex = undefined
					for i=1 to sub_count do (
						sub_tex = getSubTexmap inTexture i
						if(sub_tex != undefined) then
							setSubTexmap inTexture i (gt_createVRayTexture sub_tex)
					)
				)

				-- get the texture class id
				cls=classof inTexture

				-- if this is already a V-Ray compatible material, return the same one
				if (gt_isVRayCompatibleTex cls) then return inTexture

				-- try to convert
				if debug do
					format "  %\n" inTexture
				
				local new_tex=case cls of (
					Normal_Bump: gt_convertFrom_NormalBump inTexture 				
					Autodesk_Map: gt_convertFrom_AutodeskBitmap inTexture 
					Raytrace:gt_convertFrom_RaytraceMap inTexture 
					CoronaOutput: if inTexture.inputTexmap != undefined do inTexture.inputTexmap
					CoronaRaySwitch: gt_convertFrom_UnsupportedMap inTexture 
					CoronaFrontBack: gt_convertFrom_CoronaFrontBack inTexture
					CoronaColor: gt_convertFrom_CoronaColorTex inTexture 
					CoronaMultiMap: gt_convertFrom_CoronaMultiMapTex inTexture 
					CoronaMix: gt_convertFrom_CoronaMixTex inTexture 
					CoronaAo: gt_convertFrom_CoronaAoTex inTexture 
					CoronaNormal: gt_convertFrom_CoronaNormalTex inTexture 
					CoronaSky: gt_convertFrom_CoronaSky inTexture 
					CoronaWire: gt_convertFrom_CoronaWireTex inTexture 
					CoronaCurvature: gt_convertFrom_CoronaCurvatureTex inTexture 
					CoronaDistance: gt_convertFrom_CoronaDistance inTexture 
					CoronaRoundEdges: gt_convertFrom_CoronaRoundEdges inTexture 
					CoronaTriplanar: gt_convertFrom_CoronaTriplanar inTexture 
					CoronaBitmap: gt_convertFrom_CoronaBitmap inTexture 
					CoronaBumpConverter: gt_convertFrom_CoronaBumpConverter inTexture 
					CoronaColorCorrect: gt_convertFrom_CoronaColorCorrect inTexture 
-- 					default: testingTex inTexture
					default:  inTexture
				)

				-- if converted, add to the list of converted textures
				if new_tex!=inTexture then (
					append glsc_orig_texmaps inTexture
					append glsc_new_texmaps new_tex
				)

				-- return the new texture
				new_tex
			) -- undo off
			),			

		--convert CoronaLight texmap to VRay texture
		fn gt_coronaLightTexToVRayTex node = 
			(
				if (isValidNode node and (classof node == CoronaLight)) then (
					if debug do
						format "Converting node % texture...\n" node.name
					local tex=gt_createVRayTexture node.texmap
					gt_createSMENodeByRef node.texmap tex 
					node.texmap=tex
			-- 				format "Done\n"
				)
			),			
			
			
		-- *********************************************************************** Material Functions ***************************************************************************
		-- forward declaration
		fn gt_createVRayMtl=undefined,
			
		-- Standard Mtl
		fn gt_convertFrom_Standard origMtl = 
			(
				r=VRayMtl()
				r.name=origMtl.name
				r.effectsChannel = origMtl.effectsChannel

				r.texmap_bump=origMtl.bumpMap
				if (r.texmap_bump!=undefined) then (
					r.texmap_bump_on=origMtl.bumpMapEnable
					r.texmap_bump_multiplier=origMtl.bumpMapAmount
				)

				r.texmap_displacement=origMtl.displacementMap
				if (r.texmap_displacement!=undefined) then (
					r.texmap_displacement_on=origMtl.displacementMapEnable
					r.texmap_displacement_multiplier=origMtl.displacementMapAmount
				)

				r.diffuse=origMtl.diffuse
				r.texmap_diffuse=origMtl.diffuseMap
				if (r.texmap_diffuse!=undefined) then (
					r.texmap_diffuse_on=origMtl.diffuseMapEnable
					r.texmap_diffuse_multiplier=origMtl.diffuseMapAmount
				)

				if (origMtl.shaderByName !="Strauss") then (
					r.reflection=origMtl.specular*origMtl.specularLevel*0.01
				)

				r.refraction=(Color 255 255 255) * (1.0 - origMtl.opacity*0.01)
				

				-- unlocked highlights and reflection glossiness no longer supported in V-Ray Next and
				-- that affects reflection and reflection_glossiness initialisation
				local vrMajorVer = 3
				if vrayVersion != undefined do (
					vrMajorVer = (vrayVersion())[1][1] as Integer
				)

				K=2.0^(origMtl.glossiness*0.01*10.0)
				if (origMtl.shaderType==5) then (
					-- phong
					r.brdf_type=0
					if vrMajorVer < 4 do r.reflection=r.reflection*2.0/(K+1.0)
				) else (
					-- blinn
					r.brdf_type=1
					K=K*4
					if vrMajorVer < 4 do r.reflection=r.reflection*8.0/(K+1.0)
				)

				if vrMajorVer > 3 then (
					r.reflection_glossiness=1.0-(1.0/(K+1.0))^(1.0/3.5)
				)
				else (
					r.reflection_lockGlossiness=false
					r.hilight_glossiness=1.0-(1.0/(K+1.0))^(1.0/3.5)
					r.reflection_glossiness=1.0
				)
				
				r.option_traceReflection=true

				r.texmap_opacity=origMtl.opacityMap
				if (r.texmap_opacity!=undefined) then (
					r.texmap_opacity_on=origMtl.opacityMapEnable
					r.texmap_opacity_multiplier=origMtl.opacityMapAmount
				)

				r
			),

		-- Raytrace Mtl
		fn gt_convertFrom_Raytrace origMtl = 
			(
				r=VRayMtl()
				r.name=origMtl.name
				r.effectsChannel = origMtl.effectsChannel

				r.texmap_bump=origMtl.bumpMap
				if (r.texmap_bump!=undefined) then (
					r.texmap_bump_on=origMtl.bumpMapEnable
					r.texmap_bump_multiplier=origMtl.bumpMapAmount
				)

				r.texmap_displacement=origMtl.displacementMap
				if (r.texmap_displacement!=undefined) then (
					r.texmap_displacement_on=origMtl.displacementMapEnable
					r.texmap_displacement_multiplier=origMtl.displacementMapAmount
				)

				r.diffuse=origMtl.diffuse
				r.texmap_diffuse=origMtl.diffuseMap
				if (r.texmap_diffuse!=undefined) then (
					r.texmap_diffuse_on=origMtl.diffuseMapEnable
					r.texmap_diffuse_multiplier=origMtl.diffuseMapAmount
				)

				r.reflection=origMtl.spec__color*origMtl.specular_level*0.01*0.5

				K=2.0^(origMtl.glossiness*0.01*10.0)
				if (origMtl.shaderType==1) then (
					-- phong
					r.brdf_type=0
					r.reflection=r.reflection*2.0/(K+1.0)
				) else (
					-- blinn
					r.brdf_type=1
					K=K*4
					r.reflection=r.reflection*8.0/(K+1.0)
				)

				if (origMtl.reflect.r>r.reflection.r) then r.reflection.r=origMtl.reflect.r
				if (origMtl.reflect.g>r.reflection.g) then r.reflection.g=origMtl.reflect.g
				if (origMtl.reflect.b>r.reflection.b) then r.reflection.b=origMtl.reflect.b

				r.reflection_glossiness=1.0-(1.0/(K+1.0))^(1.0/3.5)

				if ((origMtl.reflect.r+origMtl.reflect.g+origMtl.reflect.b)/3.0<1e-6) then r.option_traceReflection=false

				r.texmap_reflection=origMtl.reflectionMap
				if (r.texmap_reflection!=undefined) then (
					r.texmap_reflection_on=origMtl.reflectionMapEnable
					r.texmap_reflection_multiplier=origMtl.reflectionMapAmount
				)

				r.refraction=origMtl.transparecy
				r.refraction_ior=origMtl.Index_of_Refraction

				r.texmap_refraction=origMtl.transparencyMap
				if (r.texmap_refraction!=undefined) then (
					r.texmap_refraction_on=origMtl.transparencyMapEnable
					r.texmap_refraction_multiplier=origMtl.transparencyMapAmount
				)

				r
			),
		
		fn gt_convertFromBlend orig_mtl = 
			(
				r=VRayBlendMtl()
				r.name=orig_mtl.name
				r.effectsChannel = orig_mtl.effectsChannel

				r.baseMtl = gt_createVRayMtl orig_mtl.map1				
				r.coatMtl[1] = gt_createVRayMtl orig_mtl.map2	 
				
				r.blend[1]=(orig_mtl.mixAmount*0.01)*(color 255 255 255)
				
				if orig_mtl.mask != undefined then (
					r.texmap_blend[1] = gt_createVRayTexture orig_mtl.mask
					if orig_mtl.maskEnabled == off then r.texmap_blend_multiplier[1]=0
				)
				r
			),

		fn gt_convertFromShellac orig_mtl = 
			(
				r=VRayBlendMtl()
				r.name=orig_mtl.name
				r.effectsChannel = orig_mtl.effectsChannel
				r.additiveMode=true

				r.baseMtl = gt_createVRayMtl orig_mtl.shellacMtl1
				r.coatMtl[1] = gt_createVRayMtl orig_mtl.shellacMtl2

				r.blend[1]=(orig_mtl.shellacColorBlend*0.01)*(color 255 255 255)

				r
			),

		fn gt_convertFromDoubleSided orig_mtl = 
			(
				r=VRay2SidedMtl()
				r.name=orig_mtl.name
				r.effectsChannel = orig_mtl.effectsChannel

				r.frontMtl = gt_createVRayMtl orig_mtl.material1
				r.backMtl = gt_createVRayMtl orig_mtl.material2
				r.translucency=orig_mtl.translucency*0.01*(color 255 255 255)
				r.backMtlOn=orig_mtl.map2Enabled

				r
			),
	
		-- Physical Mtl
		fn gt_convertFrom_PhysicalMaterial im =
			(
				msgString = ("Working on " + im.name + "\n")
				om = VRayMtl()
				om.name = im.name
				om.effectsChannel = im.effectsChannel

				-- Diffuse

				if im.base_weight_map != undefined then
				(
					om.diffuse=im.base_color --No need to mult by the float weight as it's overridden by the map
					if im.base_color_map != undefined then
					(
						om.texmap_diffuse = gt_vrayMultiplyRGBWeight im.base_color_map im.base_weight_map "BT.709" -- Alternatively, "BT.601"
					)
					else
					(
						om.texmap_diffuse = gt_vrayMultiplyRGBWeight (VrayColor name:"diffuse_color" color:im.base_color) im.base_weight_map "BT.709" -- Alternatively, "BT.601"
					)
				)
				else
				(
					om.diffuse=im.base_weight*im.base_color --Mult by the weight float as no map overrides it
					om.texmap_diffuse = im.base_color_map
				)
				om.texmap_diffuse_on = im.base_color_map_on

				om.diffuse_roughness = im.diff_roughness
				om.texmap_roughness = im.diff_rough_map
				om.texmap_roughness_on = im.diff_rough_map_on

				-- Reflection
				if im.reflectivity_map != undefined then
				(
					om.reflection=im.refl_color
					if im.refl_color_map != undefined then
					(
						om.texmap_reflection = gt_vrayMultiplyRGBWeight im.refl_color_map im.reflectivity_map "BT.709"
					)
					else
					(
						om.texmap_reflection = gt_vrayMultiplyRGBWeight (VrayColor name:"reflection_color" color:im.refl_color) im.reflectivity_map "BT.709"
					)
				)
				else
				(
					om.reflection = im.reflectivity*im.refl_color
					om.texmap_reflection = im.refl_color_map
				)

				om.texmap_reflection_on=im.refl_color_map_on

				om.brdf_useRoughness = not im.roughness_inv -- Matches BRDF use of gloss/rough
				om.reflection_glossiness = im.roughness
				om.texmap_reflectionGlossiness = im.roughness_map
				om.texmap_reflectionGlossiness_on = im.roughness_map_on

				om.reflection_metalness = im.metalness
				om.texmap_metalness = im.metalness_map
				om.texmap_metalness_on = im.metalness_map_on

				om.reflection_fresnel = true -- Forced through. there is no way to mimic the custom curve mode anyway.
				om.reflection_IOR = im.trans_ior -- IOR needs forcing too, the Physmat has only one.
				om.texmap_reflectionIOR = im.trans_ior_map
				om.texmap_reflectionIOR_on = im.trans_ior_map_on
				om.option_glossyFresnel = off -- Otherwise it won't match

				-- Refraction
				if (im.transparency != 0 or im.transparency_map != undefined ) and --verify transparency is actually active.
				(im.trans_color!=black or im.trans_color_map!=undefined ) do
				(
					if im.transparency_map != undefined then --weight is textured
					(
						om.refraction=im.trans_color --No need to mult by the float weight as it's overridden by the map
						if im.trans_color_map != undefined then
						(
							om.texmap_refraction = gt_vrayMultiplyRGBWeight im.trans_color_map im.transparency_map "BT.709"
						)
						else
						(
							om.texmap_refraction = gt_vrayMultiplyRGBWeight (VrayColor name:"refraction_color" color:im.trans_color) im.transparency_map "BT.709"
						)
					)
					else
					(
						om.refraction=im.transparency*im.trans_color
						om.texmap_refraction=im.trans_color_map
					)
					om.texmap_refraction_on = im.trans_color_map_on

					-- Roughness can either use the REFLECTION controls (maps, colors and weights) or the REFRACTION controls.
					-- If it uses the reflection controls, we just need to duplicate the maps, the BRDF mode is all set.
					-- If it uses its own controls, we need to verify its BRDF is set to either gloss or rough, and invert maps as needed (VRayMtl has one option for this, physmat has two.)
					if im.trans_roughness_lock then --use REFLECTION controls
					(
						om.refraction_glossiness = im.roughness
						om.texmap_refractionGlossiness = im.roughness_map
						om.texmap_refractionGlossiness_on = im.roughness_map_on
					)
					else -- Use REFRACTION controls
					(
						if im.trans_roughness_inv == im.roughness_inv then -- Matches reflection, copy over
						(
							om.refraction_glossiness = im.trans_roughness
							om.texmap_refractionGlossiness = im.trans_rough_map
						)
						else -- Refraction uses roughness, use reciprocal
						(
							om.refraction_glossiness = 1-im.trans_roughness
							if im.trans_rough_map != undefined do om.texmap_refractionGlossiness = gt_vrayInvertTexture im.trans_rough_map false
						)
						om.texmap_refractionGlossiness_on = im.trans_rough_map_on
					)
				)

				-- Luminance/(1500.0/pi) = scene units multiplier
				if (im.emission != 0 or im.emission_map != undefined)  do
				(
					if (im.emit_color != black or im.emit_color_map != undefined ) do
					(
						if im.emission_map != undefined then --weight is textured
						(
							om.selfIllumination=im.emit_color
							if im.emit_color_map != undefined then
							(
								om.texmap_self_illumination = gt_vrayMultiplyRGBWeight im.emit_color_map im.emission_map "BT.709"
							)
							else -- We use a vraycolormap with the emission color instead.
							(
								om.texmap_self_illumination = gt_vrayMultiplyRGBWeight (vrayColor name:"selfillum_color" color:im.emit_color) im.emission_map "BT.709"
							)
						)
						else
						(
							om.selfIllumination=im.emission*im.emit_color
							om.texmap_self_illumination=im.emit_color_map
						)
						om.selfIllumination_multiplier = im.emit_luminance / (1500.0/pi)
						om.texmap_self_illumination_on = im.emit_color_map_on
					)
				)

				-- Anisotropy is a skip, broken in the physmat
				-- Still warn the user
				(
					if im.anisotropy != 1.0 or anisotropy_map != undefined do
					(
						msgString += "\t\"Anisotropy\" isn't supported: Skipping\n"
					)
				)

				-- Special maps section:
				-- Bump
				om.texmap_bump = im.bump_map
				om.texmap_bump_multiplier = 100*im.bump_map_amt
				om.texmap_bump_on = im.bump_map_on

				-- Displacement
				om.texmap_displacement = im.displacement_map
				om.texmap_displacement_multiplier = 100*im.displacement_map_amt
				om.texmap_displacement_on = im.displacement_map_on

				-- Opacity

				om.texmap_opacity_on = im.cutout_map_on
				om.texmap_opacity = im.cutout_map
				-- Change to clip mode
				if im.cutout_map != undefined do om.option_opacityMode = 1

				-- Coat
				om.coat_amount = im.coating
				om.texmap_coat_amount = im.coat_map
				om.texmap_coat_amount_on = im.coat_map_on
				om.coat_color = im.coat_color
				om.texmap_coat_color = im.coat_color_map
				om.texmap_coat_color_on = im.coat_color_map_on
				om.coat_glossiness = if im.coat_roughness_inv then im.coat_roughness else 1-im.coat_roughness
				if im.coat_roughness_inv then
				(
					om.texmap_coat_glossiness = im.coat_rough_map --if undefined it's no issue
				)
				else
				(
					if im.coat_rough_map!= undefined do gt_vrayInvertTexture im.coat_rough_map  false -- If undefined it generates an erring texture
				)
				om.texmap_coat_glossiness_on = im.coat_rough_map_on
				om.coat_ior = im.coat_ior

				om.texmap_coat_bump_multiplier = 100*im.clearcoat_bump_map_amt
				om.texmap_coat_bump_on = im.coat_bump_map_on
				om.texmap_coat_bump = im.coat_bump_map

				-- If there is translucency, wrap the material into a VRay2SidedMtl
				om.option_reflectOnBack = true
				if im.thin_walled then -- TRANSLUCENCY
				(
					if im.scattering != 0.0 or im.scattering_map != undefined then -- We have translucency, thin-walled was valid and translatable
					(
						om.option_reflectOnBack = false

						m = VRay2SidedMtl frontMtl:om name:"vray_translucency_wrapper"
						m.mult_by_front_diffuse = true

						if im.scattering_map != undefined then -- Weight is textured
						(
							m.translucency=im.sss_color -- No need to mult by the float weight as it's overridden by the map
							if im.sss_color_map != undefined then
							(
								m.texmap_translucency = gt_vrayMultiplyRGBWeight im.sss_color_map im.scattering_map "BT.709"
							)
							else
							(
								m.texmap_translucency = gt_vrayMultiplyRGBWeight (vrayColor name:"translucency_color" color:im.sss_color) im.scattering_map "BT.709"
							)
						)
						else
						(
							m.translucency=im.scattering*im.sss_color
							m.texmap_translucency=im.sss_color_map
						)
						om = m --Replace the original material with the wrapped.
					)
					else
					(
						MsgString += "\t\"Thin-Walled\" isn't supported, and no translucency was found: Skipping\n"
					)
				)
				else -- SSS
				(
					-- Skip it for now. Blend with fastSSS2?
					MsgString += "\t\"SSS\" isn't supported: Skipping\n"
				)
				if debug do
					print MsgString
				
				om
			),

		-- Architectural mtl
		fn gt_convertFrom_Architectural orig_mtl = 
			(
				r=VRayMtl()
				r.name=orig_mtl.name
				r.effectsChannel = orig_mtl.effectsChannel

				-- diffuse component
				r.diffuse=orig_mtl.diffuse
				r.texmap_diffuse=orig_mtl.diffuseMap
				if (r.texmap_diffuse!=undefined) then (
					r.texmap_diffuse_on=orig_mtl.diffuseMapEnable
					r.texmap_diffuse_multiplier=orig_mtl.diffuseMapAmount
				)

				-- specular component
				r.reflection=(color 255 255 255)
				r.texmap_reflection=orig_mtl.shininessMap
				r.reflection_fresnel=true
				r.reflection_glossiness=(orig_mtl.shininess*0.01)*0.25+0.75

				-- refraction
				r.refraction=orig_mtl.transparency*0.01*(color 255 255 255)
				r.refraction_ior=orig_mtl.ior

				-- if there is luminance, wrap the material into a VRayBlendMtl and add a VRayLightMtl on top of it
				if orig_mtl.luminance>0 then (
					local l=VRayLightMtl()
					l.color=orig_mtl.diffuse
					l.multiplier=orig_mtl.luminance
					l.opacity_texmap=VRayColor()
					l.opacity_texmap.color=(color 0 0 0)

					local m=VRayBlendMtl()
					m.baseMtl=r
					m.coatMtl[1]=l
					m.blend[1]=(color 255 255 255)

					r=m
				)

				-- if there is translucency, wrap the material into a VRay2SidedMtl
				if orig_mtl.translucency>0 then (
					local m=VRay2SidedMtl()
					m.frontMtl=r
					m.translucency=orig_mtl.translucency*0.01*(color 255 255 255)
					r=m
				)

				r
			),
		
		-- CoronaMtl
		fn gt_convertFrom_CoronaMtl orig_mtl =
			(
-- 				format " gt_convertFrom_CoronaMtl IN mtl: % \n" orig_mtl
				
				r=VRayMtl()
				r.name=orig_mtl.name
				r.effectsChannel = orig_mtl.effectsChannel

				--Diffuse
				r.diffuse = orig_mtl.levelDiffuse * orig_mtl.colorDiffuse
				r.texmap_diffuse=orig_mtl.texmapDiffuse
				if r.texmap_diffuse != undefined then
					(
					r.texmap_diffuse_on = orig_mtl.texmapOnDiffuse
					r.texmap_diffuse_multiplier = orig_mtl.mapAmountDiffuse
					)
					
				--Reflection
				r.reflection = orig_mtl.levelReflect  * orig_mtl.colorReflect
				r.texmap_reflection = orig_mtl.texmapReflect
				if r.texmap_reflection != undefined then
					(
					r.texmap_reflection_on = orig_mtl.texmapOnReflect
					r.texmap_reflection_multiplier = orig_mtl.mapamountReflect
					)
					
				r.reflection_fresnel = true
				if orig_mtl.fresnelIor == orig_mtl.Ior then
					(
						r.reflection_lockIOR = true
						r.reflection_ior = orig_mtl.ior
						r.refraction_ior = orig_mtl.ior
					)else(
						r.reflection_lockIOR = false
						r.reflection_ior = orig_mtl.fresnelIor
						)
				
				r.texmap_reflectionIOR = orig_mtl.texmapFresnelIor
				if r.texmap_reflectionIOR != undefined then
					(
						r.texmap_reflectionIOR_on = orig_mtl.texmapOnFresnelIor
						r.texmap_reflectionIOR_multiplier = orig_mtl.mapamountFresnelIor
					)
					
				r.reflection_glossiness = orig_mtl.reflectGlossiness
				r.texmap_reflectionGlossiness = orig_mtl.texmapReflectGlossiness
				if r.texmap_reflectionGlossiness != undefined then
					(
						r.texmap_reflectionGlossiness_on = orig_mtl.texmapOnReflectGlossiness
						r.texmap_reflectionGlossiness_multiplier = orig_mtl.mapamountReflectGlossiness
					)
					
				--Anisotropy
				r.anisotropy = orig_mtl.anisotropy
				r.texmap_anisotropy = orig_mtl.texmapReflectAnisotropy
				if r.texmap_anisotropy != undefined then
					(
						r.texmap_anisotropy_on = orig_mtl.texmapOnAnisotropy
						r.texmap_anisotropy_multiplier = orig_mtl.mapamountAnisotropy
					)
				r.anisotropy_rotation = orig_mtl.anisotropyRotation
				r.texmap_anisotropy_rotation = orig_mtl.texmapReflectAnisotropyRotation
				if r.texmap_anisotropy_rotation != undefined then
					(
					r.texmap_anisotropy_rotation_on = orig_mtl.texmapOnAnisotropyRotation
					r.texmap_anisotropy_rotation_multiplier = orig_mtl.mapamountAnisotropyRotation
					)
				
				--Refraction
				r.refraction = orig_mtl.levelRefract*orig_mtl.colorRefract
				r.texmap_refraction = orig_mtl.texmapRefract
				if r.texmap_refraction != undefined then
					(
						r.texmap_refraction_on = orig_mtl.texmapOnRefract
						r.texmap_refraction_multiplier = orig_mtl.mapamountRefract
					)
					
				r.refraction_ior = orig_mtl.ior
				r.texmap_refractionIOR = orig_mtl.texmapIor
				if r.texmap_refractionIOR != undefined then
					(
						r.texmap_refractionIOR_on = orig_mtl.texmapOnIor
						r.texmap_refractionIOR_multiplier = orig_mtl.mapamountIor
					)
					
				r.refraction_glossiness = orig_mtl.refractGlossiness
				r.texmap_refractionGlossiness = orig_mtl.texmapRefractGlossiness
				if r.texmap_refractionGlossiness != undefined then
					(
						r.texmap_refractionGlossiness_on = orig_mtl.texmapOnRefractGlossiness
						r.texmap_refractionGlossiness_multiplier = orig_mtl.mapamountRefractGlossiness
					)
					
				r.refraction_affectShadows = not orig_mtl.caustics
				if  orig_mtl.thin do r.refraction_ior = 1.0
					
				--Opacity
				r.texmap_opacity = orig_mtl.texmapOpacity
				if r.texmap_opacity != undefined then(
					r.texmap_opacity_on = orig_mtl.texmapOnOpacity
					r.texmap_opacity_multiplier = orig_mtl.mapamountOpacity
					)

				--Bumpmap
				r.texmap_bump = orig_mtl.texmapBump
				if r.texmap_bump != undefined then(
					r.texmap_bump_on = orig_mtl.texmapOnBump
					r.texmap_bump_multiplier = 100.0*orig_mtl.mapamountBump
					)
					
				--Absorption
				r.refraction_fogMult  = if orig_mtl.absorptionDistance == 0.0 then 0.0 else 1.0/orig_mtl.absorptionDistance
				r.refraction_fogColor  = orig_mtl.absorptionColor
				
			--Displacement
			r.texmap_displacement = orig_mtl.texmapDisplace
				
			if r.texmap_displacement != undefined then
				(
					r.texmap_displacement_on = orig_mtl.texmapOnDisplacement
					r.texmap_displacement_multiplier = (orig_mtl.displacementMaximum - orig_mtl.displacementMinimum)
				)
				
			--Self illumination
			r.selfIllumination_multiplier = orig_mtl.levelSelfIllum
			r.selfIllumination = if orig_mtl.levelSelfIllum == 0.0 then 
				black 
			else 
				orig_mtl.colorSelfIllum
			r.texmap_self_illumination = orig_mtl.texmapSelfIllum
			if r.texmap_self_illumination != undefined then(
				r.texmap_self_illumination_on = orig_mtl.texmapOnSelfIllum
				r.texmap_self_illumination_multiplier = orig_mtl.mapamountSelfIllum
				)				
				
			-- Translucency
			if gui_CoronaToVrayConverter.chk_2sideMat.checked then
				(
				if orig_mtl.levelTranslucency > 0.0 then
					(
						newVRay2SidedMtl = VRay2SidedMtl name:orig_mtl.name
						newVRay2SidedMtl.frontMtl = r

						if orig_mtl.texmapTranslucency == undefined then(
							newVRay2SidedMtl.texmap_translucency_multiplier = 100.0*orig_mtl.levelTranslucency
							newVRay2SidedMtl.Translucency = orig_mtl.colorTranslucency
						)
						if orig_mtl.texmapTranslucency != undefined and orig_mtl.texmapOnTranslucency == true then(
							theMix = Mix map1: orig_mtl.texmapTranslucency color2: orig_mtl.colorTranslucency mixAmount: orig_mtl.mapamountTranslucency
							newVRay2SidedMtl.texmap_translucency_multiplier = 100.0*orig_mtl.levelTranslucency
							newVRay2SidedMtl.Translucency = white
							newVRay2SidedMtl.texmap_translucency = theMix
						)
						return newVRay2SidedMtl
					) else 
					return r				
				)else (
					if orig_mtl.texmapTranslucency != undefined then
						(
							r.refraction_thinWalled = on
							r.texmap_translucent = orig_mtl.texmapTranslucency 
							r.texmap_translucent_multiplier = orig_mtl.mapamountTranslucency
							r.translucency_on = 6 -- SSS
							r.translucency_amount = orig_mtl.levelTranslucency 
							r.translucency_color = orig_mtl.colorTranslucency
						)
						
				return r
						
				)
				
			),
			
		-- CoronaLayeredMtl
		fn gt_convertFrom_CoronaLayeredMtl orig_mtl = 
			(
				r=VRayBlendMtl()
				r.name=orig_mtl.name
				r.effectsChannel = orig_mtl.effectsChannel

				r.baseMtl = orig_mtl.baseMtl
				for iLayer = 1 to orig_mtl.layers.count - 1 do (
					r.coatMtl[iLayer] = orig_mtl.layers[iLayer]
					r.texmap_blend[iLayer] = orig_mtl.mixmaps[iLayer]
					resVal = orig_mtl.amounts[iLayer] * 255.0 /100.0
					r.blend[iLayer] = (color resVal resVal resVal)
					if hasProperty orig_mtl #maskAmounts then (
						r.texmap_blend_multiplier[iLayer] = orig_mtl.maskAmounts[iLayer]
					)
					if hasProperty r #coatMtl_enable then(  -- Vray 3.3
						if orig_mtl.layersOn[iLayer] = false then r.coatMtl_enable[iLayer] == false
					)
				)

				return r
			),
		
		-- CoronaLightMtl
		fn gt_convertFrom_CoronaLightMtl orig_mtl = 
			(
				r=VRayLightMtl()
				r.name=orig_mtl.name
				r.effectsChannel = orig_mtl.effectsChannel

				r.color = orig_mtl.color
				r.multiplier = orig_mtl.intensity
				r.directLight_on = orig_mtl.emitLight

				if orig_mtl.texmap != undefined then(
					r.texmap_on = orig_mtl.texmapOn
					r.texmap = orig_mtl.texmap
				)
				if orig_mtl.opacityTexmap != undefined then(
					r.opacity_texmap_on = orig_mtl.opacityTexmapOn
					r.opacity_texmap = orig_mtl.opacityTexmap
				)
				return r
			),

		-- CoronaRaySwitchMtl
		fn gt_convertFrom_CoronaRaySwitchMtl orig_mtl = 
			(
				r = VRayOverrideMtl()
				r.name=orig_mtl.name
				r.effectsChannel = orig_mtl.effectsChannel

				r.baseMtl = orig_mtl.directMtl
				r.giMtl = orig_mtl.giMtl
				r.reflectMtl = orig_mtl.reflectMtl
				r.refractMtl = orig_mtl.refractMtl

				return r
			),

		-- CoronaVolumeMtl
		fn gt_convertFrom_CoronaVolumeMtl orig_mtl =
			(
				vr=VRayScatterVolume()
				vr.name=orig_mtl.name
				vr.effectsChannel = orig_mtl.effectsChannel

				vr.overall_color = orig_mtl.absorptionColor
				vr.sss_color = orig_mtl.scatteringColor
				vr.scatter_color = orig_mtl.absorptionColor
				vr.phase_function = orig_mtl.scatteringDirectionality
				vr.scatter_radius = orig_mtl.absorptionDist
				if orig_mtl.singleBounce then vr.bounces = 1

				if orig_mtl.absorptionColorTexmap != undefined and orig_mtl.absorptionColorTexmapOn == true then (
					vr.texmap_overall_color = orig_mtl.absorptionColorTexmap
					vr.texmap_scatter_color = orig_mtl.absorptionColorTexmap
				)

				if orig_mtl.scatteringColorTexmap != undefined and orig_mtl.scatteringColorTexmapOn == true then (
					vr.texmap_sss_color = orig_mtl.scatteringColorTexmap
				)

				if orig_mtl.emissionDist > 0 then (
					vrLightMtl=VRayLightMtl()
					vrLightMtl.name = "vraylight "+orig_mtl.name
					vrLightMtl.color = orig_mtl.emissionColor
					vrLightMtl.multiplier = (1.0f / orig_mtl.emissionDist)

					vrColorMap = VRayColor()
					vrColorMap.color = (Color 0 0 0)
					vrLightMtl.opacity_texmap = vrColorMap

					vrBlendMtl = VRayBlendMtl()
					vrBlendMtl.name = "vrayblend "+orig_mtl.name
					vrBlendMtl.baseMtl = vr
					vrBlendMtl.coatMtl[1] = vrLightMtl
					return vrBlendMtl
				)

				return vr
			),

		-- NOT USED
		-- CoronaShadowCatcherMtl
		fn gt_convertFrom_CoronaShadowCatcherMtl orig_mtl = 
			(
				vrwrap=VRayMtlWrapper()
				vrwrap.name=orig_mtl.name
				vrwrap.effectsChannel = orig_mtl.effectsChannel

				vrwrap.matteSurface = true
				vrwrap.alphaContribution = case orig_mtl.alphaMode of (
					1: 1.0f
					2: -1.0f
					3: 0.0f
					default: -1.0f
				)
				vrwrap.matte_shadows = true
				vrwrap.matte_shadowsAffectAlpha = true
				vrwrap.matte_shadowBrightness = orig_mtl.shadowAmount

				vr=VRayMtl()
				vr.diffuse = orig_mtl.backplateColor
				if orig_mtl.backplateTexmap != undefined then (
					vr.texmap_diffuse = orig_mtl.backplateTexmap
					vr.texmap_diffuse_on = true
				)

				--Reflection
				vr.reflection = orig_mtl.reflectLevel  * orig_mtl.colorReflect
				vr.texmap_reflection = orig_mtl.reflectColorTexmap

				vr.reflection_fresnel = true
				vr.reflection_lockIOR = false
				vr.reflection_ior = orig_mtl.reflectIor
				vr.reflection_glossiness = orig_mtl.reflectGlossiness

				vr.texmap_bump = orig_mtl.bump
				if vr.texmap_bump != undefined then(
					vr.texmap_bump_on = true
					vr.texmap_bump_multiplier = 100.0*orig_mtl.bumpAmount
				)

				vrwrap.baseMtl = vr
				return vrwrap
			),

		-- CoronaHairMtl
		fn gt_convertFrom_CoronaHairMtl orig_mtl = 
			(
				vr=VRayHairNextMtl()
				vr.name=orig_mtl.name
				vr.effectsChannel = orig_mtl.effectsChannel

				vr.melanin = orig_mtl.melanin
				vr.random_melanin = orig_mtl.randomizeMelanin
				vr.texmap_melanin_on = orig_mtl.texmapOnMelanin
				vr.texmap_melanin_multiplier = orig_mtl.mapamountMelanin
				vr.texmap_melanin = orig_mtl.texmapMelanin
				vr.pheomelanin = orig_mtl.pheomelanin

				vr.dye_color = orig_mtl.levelTint * orig_mtl.colorTint
				vr.texmap_dye_color_on = orig_mtl.texmapOnTint
				vr.texmap_dye_color_multiplier = orig_mtl.mapamountTint
				if (orig_mtl.levelTint == 1) then (
					vr.texmap_dye_color = orig_mtl.texmapTint
				) else if (orig_mtl.levelTint == 0) then (
					vr.texmap_dye_color_on = off
					vr.texmap_dye_color = orig_mtl.texmapTint
				) else (
					outTexDye = output()
					outTexDye.map1 = orig_mtl.texmapTint
					outTexDye.output.output_amount = orig_mtl.levelTint
					vr.texmap_dye_color = outTexDye
				)

				vr.opacity = orig_mtl.levelOpacity * orig_mtl.colorOpacity
				vr.texmap_opacity_on = orig_mtl.texmapOnOpacity
				vr.texmap_opacity_multiplier = orig_mtl.mapamountOpacity
				if (orig_mtl.levelOpacity == 1) then (
					vr.texmap_opacity = orig_mtl.texmapOpacity
				) else if (orig_mtl.levelOpacity == 0) then (
					vr.texmap_opacity_on = off
					vr.texmap_opacity = orig_mtl.texmapOpacity
				) else (
					outTexOpacity = output()
					outTexOpacity.map1 = orig_mtl.texmapOpacity
					outTexOpacity.output.output_amount = orig_mtl.levelOpacity
					vr.texmap_opacity = outTexOpacity
				)

				vr.glossiness = orig_mtl.glossiness
				vr.texmap_glossiness_on = orig_mtl.texmapOnGlossiness
				vr.texmap_glossiness_multiplier = orig_mtl.mapamountGlossiness
				vr.texmap_glossiness = orig_mtl.texmapGlossiness
				vr.primary_glossiness_boost = orig_mtl.colorlessGlossiness
				vr.texmap_primary_glossiness_boost_on = orig_mtl.texmapOnColorlessGlossiness
				vr.texmap_primary_glossiness_boost_multiplier = orig_mtl.mapamountColorlessGlossiness
				vr.texmap_primary_glossiness_boost = orig_mtl.texmapColorlessGlossiness
				vr.highlight_shift = orig_mtl.highlightShift

				vr.secondary_tint = orig_mtl.coloredSpecular
				vr.texmap_secondary_tint = orig_mtl.texmapColoredSpecular
				vr.texmap_secondary_tint_on = orig_mtl.texmapOnColoredSpecular
				vr.texmap_secondary_tint_multiplier = orig_mtl.mapamountColoredSpecular

				vr.primary_tint = orig_mtl.colorlessSpecular
				vr.texmap_primary_tint = orig_mtl.texmapColorlessSpecular
				vr.texmap_primary_tint_on = orig_mtl.texmapOnColorlessSpecular
				vr.texmap_primary_tint_multiplier = orig_mtl.mapamountColorlessSpecular

				vr.transmission_tint = orig_mtl.transmission
				vr.texmap_transmission_tint = orig_mtl.texmapTransmission
				vr.texmap_transmission_tint_on = orig_mtl.texmapOnTransmission
				vr.texmap_transmission_tint_multiplier = orig_mtl.mapamountTransmission

				vr.diffuse_amount = orig_mtl.levelDiffuse
				vr.diffuse_color = orig_mtl.colorDiffuse
				vr.texmap_diffuse_color = orig_mtl.texmapDiffuse
				vr.texmap_diffuse_color_on = orig_mtl.texmapOnDiffuse
				vr.texmap_diffuse_color_multiplier = orig_mtl.mapAmountDiffuse

				vr.softness = orig_mtl.softness
				vr.texmap_softness = orig_mtl.texmapSoftness
				vr.texmap_softness_on = orig_mtl.texmapOnSoftness
				vr.texmap_softness_multiplier = orig_mtl.mapamountSoftness

				vr.ior = orig_mtl.ior
				vr.texmap_ior = orig_mtl.texmapIor
				vr.texmap_ior_on = orig_mtl.texmapOnIor
				vr.texmap_ior_multiplier = orig_mtl.mapamountIor

				vr
			),

		-- CoronaSkinMtl
		fn gt_convertFrom_CoronaSkinMtl orig_mtl = 
			(
				vr=VRayALSurfaceMtl()
				vr.name=orig_mtl.name
				vr.effectsChannel = orig_mtl.effectsChannel

				vr.sss_density_scale = orig_mtl.sssRadiusScale

				vr.opacity = orig_mtl.opacityLevel*orig_mtl.opacityColor
				if (orig_mtl.texmapOpacityColor != undefined) do (
					vr.opacity_texture_on = orig_mtl.texmapOnOpacityColor
					vr.opacity_texture_multiplier = orig_mtl.mapamountOpacityColor
					if (orig_mtl.opacityLevel == 1) then (
						vr.opacity_texture = orig_mtl.texmapOpacityColor
					) else if (orig_mtl.opacityLevel == 0) then (
						vr.opacity_texture_on = off
						vr.opacity_texture = orig_mtl.texmapOpacityColor
					) else (
						outTexOpacity = output()
						outTexOpacity.map1 = orig_mtl.texmapOpacityColor
						outTexOpacity.output.output_amount = orig_mtl.opacityLevel
						vr.opacity_texture = outTexOpacity
					)
				)

				vr.displacement_texture = orig_mtl.texmapDisplace
				vr.displacement_texture_on = orig_mtl.texmapOnDisplace
				--vr.displacement_texture_multiplier = displaceMinimum?, displaceWaterLevel?, displaceMaximum?

				vr.bump_texture = orig_mtl.texmapBump
				vr.bump_texture_on = orig_mtl.texmapOnBump
				vr.bump_texture_multiplier = orig_mtl.mapamountBump

				vr.diffuse_color = orig_mtl.skinColor
				vr.diffuse_strength = orig_mtl.skinLevel
				vr.diffuse_color_texture = orig_mtl.texmapSkinColor
				vr.diffuse_color_texture_on = orig_mtl.texmapOnSkinColor
				vr.diffuse_color_texture_multiplier = orig_mtl.mapamountSkinColor
				vr.sss_mix = orig_mtl.sssAmount
				vr.sss_mix_texture = orig_mtl.texmapSssAmount
				vr.sss_mix_texture_on = orig_mtl.texmapOnSssAmount
				vr.sss_mix_texture_multiplier = orig_mtl.mapamountSssAmount
				vr.sss_mode = orig_mtl.sssMode
				if orig_mtl.sssMode == 0 then vr.sss_mode = 1
				else vr.sss_mode = 0

				--SSS1
				vr.sss1_color = orig_mtl.sss1Color
				vr.sss1_color_texture = orig_mtl.texmapSss1Color
				vr.sss1_color_texture_on = orig_mtl.texmapOnSss1Color
				vr.sss1_color_texture_multiplier = orig_mtl.mapamountSss1Color
				vr.sss1_weight = orig_mtl.sss1Weight
				vr.sss1_weight_texture = orig_mtl.texmapSss1Weight
				vr.sss1_weight_texture_on = orig_mtl.texmapOnSss1Weight
				vr.sss1_weight_texture_multiplier = orig_mtl.mapamountSss1Weight
				vr.sss1_radius = orig_mtl.sss1Radius
				vr.sss1_radius_texture = orig_mtl.texmapSss1Radius
				vr.sss1_radius_texture_on = orig_mtl.texmapOnSss1Radius
				vr.sss1_radius_texture_multiplier = orig_mtl.mapamountSss1Radius

				--SSS2
				vr.sss2_color = orig_mtl.sss2Color
				vr.sss2_color_texture = orig_mtl.texmapSss2Color
				vr.sss2_color_texture_on = orig_mtl.texmapOnSss2Color
				vr.sss2_color_texture_multiplier = orig_mtl.mapamountSss2Color
				vr.sss2_weight = orig_mtl.sss2Weight
				vr.sss2_weight_texture = orig_mtl.texmapSss2Weight
				vr.sss2_weight_texture_on = orig_mtl.texmapOnSss2Weight
				vr.sss2_weight_texture_multiplier = orig_mtl.mapamountSss2Weight
				vr.sss2_radius = orig_mtl.sss2Radius
				vr.sss2_radius_texture = orig_mtl.texmapSss2Radius
				vr.sss2_radius_texture_on = orig_mtl.texmapOnSss2Radius
				vr.sss2_radius_texture_multiplier = orig_mtl.mapamountSss2Radius

				--SSS3
				vr.sss3_color = orig_mtl.sss3Color
				vr.sss3_color_texture = orig_mtl.texmapSss3Color
				vr.sss3_color_texture_on = orig_mtl.texmapOnSss3Color
				vr.sss3_color_texture_multiplier = orig_mtl.mapamountSss3Color
				vr.sss3_weight = orig_mtl.sss3Weight
				vr.sss3_weight_texture = orig_mtl.texmapSss3Weight
				vr.sss3_weight_texture_on = orig_mtl.texmapOnSss3Weight
				vr.sss3_weight_texture_multiplier = orig_mtl.mapamountSss3Weight
				vr.sss3_radius = orig_mtl.sss3Radius
				vr.sss3_radius_texture = orig_mtl.texmapSss3Radius
				vr.sss3_radius_texture_on = orig_mtl.texmapOnSss3Radius
				vr.sss3_radius_texture_multiplier = orig_mtl.mapamountSss3Radius

				--Reflection1
				vr.reflect1_color = orig_mtl.reflection1Color
				vr.reflect1_color_texture = orig_mtl.texmapReflection1Color
				vr.reflect1_color_texture_on = orig_mtl.texmapOnReflection1Color
				vr.reflect1_color_texture_multiplier = orig_mtl.mapamountReflection1Color
				vr.reflect1_strength = orig_mtl.reflection1Level
				vr.reflect1_roughness = 1 - orig_mtl.reflection1Glossiness
				if (orig_mtl.texmapReflection1Glossiness != undefined) do (
					outRefl1RoughTex = output()
					outRefl1RoughTex.map1 = orig_mtl.texmapReflection1Glossiness
					outRefl1RoughTex.output.invert = on
					vr.reflect1_roughness_texture = outRefl1RoughTex
					vr.reflect1_roughness_texture_on = orig_mtl.texmapOnReflection1Glossiness
					vr.reflect1_roughness_texture_multiplier = orig_mtl.mapamountReflection1Glossiness
				)
				vr.reflect1_ior = orig_mtl.reflection1Ior
				vr.reflect1_ior_texture = orig_mtl.texmapReflection1Ior
				vr.reflect1_ior_texture_on = orig_mtl.texmapOnReflection1Ior
				vr.reflect1_ior_texture_multiplier = orig_mtl.mapamountReflection1Ior

				--Reflection2
				vr.reflect2_color = orig_mtl.reflection2Color
				vr.reflect2_color_texture = orig_mtl.texmapReflection2Color
				vr.reflect2_color_texture_on = orig_mtl.texmapOnReflection2Color
				vr.reflect2_color_texture_multiplier = orig_mtl.mapamountReflection2Color
				vr.reflect2_strength = orig_mtl.reflection2Level
				vr.reflect2_roughness = 1 - orig_mtl.reflection2Glossiness
				if (orig_mtl.texmapReflection2Glossiness != undefined) do (
					outRefl2RoughTex = output()
					outRefl2RoughTex.map1 = orig_mtl.texmapReflection2Glossiness
					outRefl2RoughTex.output.invert = on
					vr.reflect2_roughness_texture = outRefl2RoughTex
					vr.reflect2_roughness_texture_on = orig_mtl.texmapOnReflection2Glossiness
					vr.reflect2_roughness_texture_multiplier = orig_mtl.mapamountReflection2Glossiness
				)
				vr.reflect2_ior = orig_mtl.reflection2Ior
				vr.reflect2_ior_texture = orig_mtl.texmapReflection2Ior
				vr.reflect2_ior_texture_on = orig_mtl.texmapOnReflection2Ior
				vr.reflect2_ior_texture_multiplier = orig_mtl.mapamountReflection2Ior

				vr
			),
			
		--CORONA PHYSICAL material conversion
		--functions
		-- @param cpm = Corona physical material
		-- @param vrm = V-Ray material
		-- "Full" mode will use vraycomptex and other such maps to ensure the closest possible conversion result (to factor in amounts and other multipliers we do not have.)
		-- "Simple" will skip all that and just replug the maps, and set their influence.
		fn  gt_convertFrom_CoronaPhysicalMtl cpm mode=
			(
				vrm = vrayMtl name:cpm.name
				vrm.effectsChannel = cpm.effectsChannel

				--material flags
				----Metalness
				isMetal = cpm.metalnessMode as booleanClass --we need this for the complex IoR
				vrm.reflection_metalness = cpm.metalnessMode as float

				----Metalness Texture
				if cpm.metalnessTexmap != undefined do
				(
					isMetal = true
					vrm.texmap_metalness = cpm.metalnessTexmap
				)

				----Thinness
				isThin = cpm.useThinMode
				vrm.refraction_thinWalled = isThin

				----Self Illumination
				/*
				Our material's self-illumination matches color*level for lights when the level is expressed in *scene units*.
				Their materials match color*level for lights when the level is expressed in *Radiance*.
				However, when they render our lights, they take the scene units multiplier and treat it straight as if it was radiance.
				The selfillum level should be rechecked!
				*/
				vrm.selfIllumination_multiplier = cpm.selfIllumLevel
				vrm.selfIllumination = cpm.selfIllumColor
				vrm.texmap_self_illumination = cpm.selfIllumTexmap

				----gBuffer channel
				vrm.effectsChannel = if cpm.gBufferOverride >= 0 then cpm.gBufferOverride else cpm.effectsChannel -- -1 uses the normal gbuffer id, anything else overrides it.

				----Alpha mode
				----ours is automatic

				----propagate masks
				case cpm.renderElementPropagation of
				(
					0 :()
					1 :
					(
						vrm.reflection_affectAlpha = 2
					)
					2 :
					(
						vrm.refraction_affectAlpha = 2
					)
					3 :
					(
						vrm.reflection_affectAlpha = 2
						vrm.refraction_affectAlpha = 2
					)
				)

				----Roughness/Glossiness
				isRough = not (cpm.roughnessMode as booleanClass)
				vrm.brdf_useRoughness = isRough  --boolean

				----Specular Mode and IoR
				isDisney = cpm.iorMode as booleanClass --0 is default, 1 is disney.
				if isDisney then
				(
					vrm.refraction_ior = gt_getIorFromDisney cpm.baseIor
				)
				else
				(
					vrm.refraction_ior = cpm.baseIor
				)

				---- Complex IoR
				if isMetal and cpm.useComplexIor do
				(
					vrm.texmap_reflection = gt_setupComplexIoR cpm
					vrm.reflection_fresnel = off
				)

				--Shading Layers
				----Base Color
				if mode == "full" then
				(
					if cpm.baseLevel == 1.0 then --no base level is set, no need for composite maps.
					(
						vrm.diffuse =  cpm.basecolor --this is always true, regardless of metallicity.
						vrm.texmap_diffuse = cpm.baseTexmap --always copy this one. Metals will override it as needed (f.e. with complex IoR)
						if isMetal then
						(
							vrm.reflection = cpm.edgeColor
							if vrm.texmap_reflection == undefined do vrm.texmap_reflection = cpm.edgeColorTexmap --if complex IoR was set, we don't want to override the texture.
							vrm.texmap_reflection_multiplier = cpm.edgeColorMapAmount
							vrm.refraction_ior = 1.6 --Vlado's suggestion.
						)
						else
						(
							vrm.reflection = white
							vrm.refraction_ior = cpm.baseIor
							vrm.texmap_refractionIOR= cpm.baseIorTexmap
						)
					)
					else --the value is lower than 1.0, we need output maps to fix this.
					(
						vrm.diffuse = cpm.basecolor*cpm.baseLevel --this is always true, regardless of metallicity.
						if cpm.baseTexmap!= undefined do
						(
							vrm.texmap_diffuse = output map1:cpm.baseTexmap name:"Base Level Explicitation"
							vrm.texmap_diffuse.output.RGB_Level = cpm.baseLevel
						)
						if isMetal then
						(
							vrm.reflection = cpm.edgeColor*cpm.baseLevel
							if vrm.texmap_reflection == undefined and cpm.edgeColorTexmap != undefined do
							(
								vrm.texmap_reflection = output map1:cpm.edgeColorTexmap name:"Base Level Explicitation"
								vrm.texmap_reflection.output.RGB_Level = cpm.baseLevel
							)
							vrm.texmap_reflection_multiplier = cpm.edgeColorMapAmount
							vrm.refraction_ior = 1.6 --Vlado's suggestion.
						)
						else
						(
							vrm.reflection = white * cpm.baseLevel -- Convert to gray baseLevel color
							vrm.refraction_ior = cpm.baseIor
							vrm.texmap_refractionIOR= cpm.baseIorTexmap
						)
					)
				)
				else --mode is "simple"
				(
					vrm.diffuse =  cpm.basecolor --this is always true, regardless of metallicity.
					vrm.texmap_diffuse = cpm.baseTexmap --always copy this one. Metals will override it as needed (f.e. with complex IoR)
					if isMetal then
					(
						vrm.reflection = cpm.edgeColor
						if vrm.texmap_reflection == undefined do vrm.texmap_reflection = cpm.edgeColorTexmap --if complex IoR was set, we don't want to override the texture.
						vrm.texmap_reflection_multiplier = cpm.edgeColorMapAmount
						vrm.refraction_ior = 1.6 --Vlado's suggestion.
					)
					else
					(
						vrm.reflection = white
						vrm.refraction_ior = cpm.baseIor
						vrm.texmap_refractionIOR= cpm.baseIorTexmap
					)
				)

				----Base Roughness
				vrm.reflection_glossiness = cpm.baseRoughness --the V-Ray material is already set to use the right mode.
				vrm.texmap_reflectionGlossiness = cpm.baseRoughnessTexmap
				vrm.refraction_glossiness = cpm.baseRoughness
				vrm.texmap_refractionGlossiness = cpm.baseRoughnessTexmap

				----Base Bump
				vrm.texmap_bump_multiplier = gt_remapValue cpm.baseBumpMapAmount -99. 99. -1000. 1000.
				vrm.texmap_bump = cpm.baseBumpTexmap

				----Base Anisotropy
				vrm.anisotropy = cpm.baseAnisotropy
				vrm.texmap_anisotropy = cpm.baseAnisotropyTexmap
				vrm.anisotropy_rotation = cpm.baseAnisoRotation + 90
				vrm.texmap_anisotropy_rotation = cpm.baseAnisoRotationTexmap
				vrm.anisotropy_derivation = cpm.anisotropyOrientationMode
				vrm.anisotropy_channel = cpm.anisotropyOrientationUvwChannel

				----Base Translucency
				if isThin and cpm.translucencyFraction > 0.0 do
				(

					vrm.translucency_on = 6
					vrm.translucency_amount = cpm.translucencyFraction
					vrm.translucency_color = cpm.translucencyColor
					if cpm.translucencyFractionTexmap != undefined and mode == "full" then --mapped translucency amount, make explicit multiplying by translucency texture
					(
						vrm.texmap_translucent = vrayCompTex name:"textured translucency fraction explicitation" sourceA:cpm.translucencyColorTexmap sourceB:cpm.translucencyFractionTexmap operator:3
					)
					else
					(
						vrm.texmap_translucent = cpm.translucencyColorTexmap --no map, no worry.
					)
					vrm.texmap_translucent_multiplier = cpm.translucencyColorMapAmount --as this control is reused on our side, the map amount has to be set where the effect is active
					vrm.texmap_translucent_on = cpm.translucencyColorTexmapOn
				)

				----Refraction
				vrm.refraction = white * cpm.refractionAmount -- Convert to gray reflectionAmount color
				if cpm.refractionAmountTexmap != undefined and cpm.refractionAmountTexmapOn do vrm.texmap_refraction = rgb_tint map1:cpm.refractionAmountTexmap red:white green:white blue:white name:"refraction amount texture mono transform"
				vrm.refraction_affectShadows = not cpm.useCaustics
				vrm.refraction_dispersion_on = cpm.dispersionEnable
				vrm.refraction_dispersion = cpm.dispersion

				----Thin Absorption
				if isThin do
				(
					vrm.refraction_fogColor = cpm.thinAbsorptionColor
					vrm.texmap_refraction_fog = cpm.thinAbsorptionTexmap
					vrm.texmap_refraction_fog_multiplier = cpm.thinAbsorptionMapAmount
					vrm.texmap_refraction_fog_on = cpm.thinAbsorptionTexmapOn
				)

				----Opacity
				if cpm.opacityColor != white do --as we don't have a color, but just a texture, i need to do liek below.
				(
					vrm.texmap_opacity = vrayColor color:cpm.opacityColor name:"Opacity Color" rgb_multiplier:cpm.opacityLevel
				)
				if cpm.opacityTexmap != undefined do --i need this to cater for the opacity level, separate from the opacity map amount.
				(
					if mode == "full" then
					(
						vrmColorMap = vrayColor color:cpm.opacityColor name:"Opacity Color" rgb_multiplier:cpm.opacityLevel
						vrm.texmap_opacity = vrayCompTex sourceA:vrmColorMap sourceB:cpm.opacityTexmap operator:3 multiplier:(cpm.opacityMapAmount/100.0)
					)
					else --mode is "simple"
					(
						vrm.texmap_opacity = cpm.opacityTexmap
						vrm.texmap_opacity_multiplier = cpm.opacityMapAmount
					)
				)
				if cpm.opacityCutout do vrm.option_opacityMode = 1

				----Displacement
				----I'll only plug the texture as the rest of the settings aren't present on the v-ray material
				----No map influence for the CPM.
				vrm.texmap_displacement = cpm.displacementTexmap

				----Clearcoat
				vrm.coat_amount = cpm.clearcoatAmount

				vrm.texmap_coat_amount = cpm.clearcoatAmountTexmap
				vrm.coat_glossiness = cpm.clearcoatRoughness --this is properly set to rough/gloss at the top for the whole shader, pipe 1:1
				vrm.texmap_coat_glossiness = cpm.clearcoatRoughnessTexmap
				vrm.coat_ior = cpm.clearcoatIor
				vrm.texmap_coat_ior = cpm.clearcoatIorTexmap
				vrm.coat_color = cpm.clearcoatAbsorptionColor
				vrm.texmap_coat_color = cpm.clearcoatAbsorptionTexmap
				vrm.texmap_coat_bump_multiplier = gt_remapValue cpm.clearcoatBumpMapAmount -99. 99. -1000. 1000. --this hasn't got an Ui in the right place for the corona material. so odd.
				vrm.texmap_coat_bump = cpm.clearcoatBumpTexmap


				----Sheen
				/*
				Sheen has an amount and a textured amount, which we lack.
				*/

				if mode == "full" and ((cpm.sheenAmount > 0.0 and cpm.sheenAmount < 1.0) or cpm.sheenAmountTexmap != undefined) then --influence lower than 1.0, and/or textured influence, need output maps
				(
					if cpm.sheenAmountTexmap == undefined then --just the level, this is doable with outputs.
					(
						vrm.sheen_color = cpm.sheenColor * cpm.sheenAmount
						if cpm.sheenColorTexmap != undefined do
						(
							vrm.texmap_sheen = output map1:cpm.sheenColorTexmap name:"Sheen Amount Explicitation"
							vrm.texmap_sheen.output.rgb_level = cpm.sheenAmount
							vrm.texmap_sheen_multiplier = cpm.sheenColorMapAmount
						)
					)
					else --we have a texture driving amount, need a comptex
					(
						sheenAmountMap = vrayColor color:white rgb_multiplier:cpm.sheenAmount --made the float amount a color.
						monoAmountMap =  rgb_tint map1:cpm.sheenAmountTexmap red:white green:white blue:white name:"Sheen amount texture mono transform"
						globalSheenAmountMap = vrayCompTex name:"globalSheenAmountMap" sourceA:sheenAmountMap sourceB:monoAmountMap operator:3
						globalSheenAmountMap.multiplier=(cpm.sheenAmountMapAmount/100.)
						sheenColorMap = vrayColor color:cpm.sheenColor

						if cpm.sheenColorTexmap != undefined then --
						(
							sheenTexturedColour = vrayCompTex name:"textured Sheen amount explicitation" sourceA:sheenColorMap sourceB:cpm.sheenColorTexmap operator:9 multiplier:(cpm.sheenColorMapAmount/100.)
							vrm.texmap_sheen = vrayCompTex name:"globalSheenColorMap" sourceA:sheenTexturedColour sourceB:globalSheenAmountMap operator:3
						)
						else
						(
							texturedAmount = vrayCompTex name:"textured Sheen amount explicitation" sourceA:sheenColorMap sourceB:monoMap operator:3
							texturedAmount.multiplier=(cpm.sheenColorMapAmount/100.) --oddly it's 0-1 while the UI shows 0-100
							vrm.texmap_sheen = texturedAmount
						)
					)
				)
				else --influence at 1.0, or we chose simple mode. we can just plug away.
				(
					vrm.sheen_color = cpm.sheenColor * cpm.sheenAmount
					vrm.texmap_sheen = cpm.sheenColorTexmap
					vrm.texmap_sheen_multiplier = cpm.sheenColorMapAmount
				)
				vrm.sheen_glossiness = cpm.sheenRoughness
				vrm.texmap_sheen_glossiness = cpm.sheenRoughnessTexmap

				----Fog
				vrm.refraction_fogColor = cpm.volumetricAbsorptionColor
				vrm.texmap_refraction_fog = cpm.volumetricAbsorptionTexmap
				vrm.texmap_refraction_fog_multiplier = cpm.volumetricAbsorptionMapAmount
				vrm.texmap_refraction_fog_on = cpm.volumetricAbsorptionTexmapOn
				vrm.refraction_fogDepth = cpm.attenuationDistance

				----Volumetric Scattering
				if cpm.volumetricScatteringColor != black or cpm.volumetricScatteringTexmap != undefined do
				(
					vrm.translucency_on = 5 --vol. scatter
					vrm.translucency_color = cpm.volumetricScatteringColor
					vrm.texmap_translucent = cpm.volumetricScatteringTexmap
					vrm.texmap_translucent_multiplier = cpm.volumetricScatteringMapAmount
					vrm.texmap_translucent_on = cpm.volumetricScatteringTexmapOn
				)

				----SSS
				if cpm.sssAmount > 0.0 or cpm.sssAmountTexmap != undefined do
				(
					vrm.translucency_on = 6 --SSS
					vrm.translucency_amount = cpm.sssAmount
					vrm.translucency_color = cpm.sssScatterColor
					if mode == "full" and (cpm.sssAmountTexmap != undefined and cpm.sssAmountTexmapOn) then --textured influence. Make explicit.
					(
						monoMap = rgb_tint map1:cpm.sssAmountTexmap red:white green:white blue:white name:"SSS amount texture mono transform"
						vrm.texmap_translucent = vrayCompTex name:"textured SSS amount explicitation" sourceA:monoMap sourceB:cpm.sssScatterTexmap operator:3
						vrm.texmap_translucent_multiplier = cpm.sssScatterMapAmount
						vrm.texmap_translucent_on = cpm.sssScatterTexmapOn --troublesome, as the map is explicitated above.
					)
					else --mode is "simple", or no textured influence
					(
						vrm.texmap_translucent = cpm.sssScatterTexmap --no map, no worry.
						vrm.texmap_translucent_multiplier = cpm.sssScatterMapAmount
						vrm.texmap_translucent_on =	cpm.sssScatterTexmapOn
					)
					vrm.refraction_fogDepth = cpm.sssRadius * 100
					if cpm.sssRadiusTexmap != undefined and cpm.sssRadiusTexmapOn do
					(
						vrm.texmap_refraction_fog = output map1:cpm.sssRadiusTexmap name:"SSS Radiut units conversion"
						vrm.texmap_refraction_fog.output.RGB_Level = 1
						vrm.texmap_refraction_fog_multiplier = cpm.sssRadiusMapAmount
						vrm.texmap_refraction_fog_on = cpm.sssRadiusTexmapOn
					)
				)

				--extra maps
				----Environment: i plug the refraction one first, and the reflection one then, so to give it higher priority.
				----As we only have one slot, this ensures the reflection one is the one that will be chosen if both are plugged in the CPM.
				vrm.texmap_environment = cpm.bgOverrideRefractTexmap
				vrm.texmap_environment = cpm.bgOverrideReflectTexmap

				--Unconditional Map Influences
				vrm.texmap_diffuse_multiplier = cpm.baseMapAmount
				vrm.texmap_reflectionGlossiness_multiplier = cpm.baseRoughnessMapAmount
				vrm.texmap_refraction_multiplier = cpm.refractionAmountMapAmount

				vrm.texmap_self_illumination_multiplier = cpm.selfillumMapAmount
				vrm.texmap_anisotropy_multiplier = cpm.baseAnisotropyMapAmount
				vrm.texmap_anisotropy_rotation_multiplier = cpm.baseAnisoRotationMapAmount
				vrm.texmap_refractionIOR_multiplier = cpm.baseIorMapAmount
				vrm.texmap_coat_amount_multiplier = cpm.clearcoatAmountMapAmount
				vrm.texmap_coat_ior_multiplier = cpm.clearcoatIorMapAmount
				vrm.texmap_coat_color_multiplier = cpm.clearcoatAbsorptionMapAmount
				vrm.texmap_coat_glossiness_multiplier = cpm.clearcoatRoughnessMapAmount
				vrm.texmap_sheen_glossiness_multiplier = cpm.sheenRoughnessMapAmount

				--Unconditional Map Toggles
				vrm.texmap_diffuse_on = cpm.baseTexmapOn
				vrm.texmap_reflection_on = cpm.edgeColorTexmapOn
				vrm.texmap_reflectionGlossiness_on = cpm.baseRoughnessTexmapOn
				vrm.texmap_bump_on = cpm.baseBumpTexmapOn
				vrm.texmap_opacity_on = cpm.opacityTexmapOn
				vrm.texmap_displacement_on = cpm.displacementTexmapOn
				vrm.texmap_self_illumination_on = cpm.selfIllumTexmapOn
				vrm.texmap_metalness_on = cpm.metalnessTexmapOn
				vrm.texmap_anisotropy_on = cpm.baseAnisotropyTexmapOn
				vrm.texmap_anisotropy_rotation_on = cpm.baseAnisoRotationTexmapOn
				vrm.texmap_refractionIOR_on = cpm.baseIorTexmapOn
				vrm.texmap_coat_amount_on = cpm.clearcoatAmountTexmapOn
				vrm.texmap_coat_glossiness_on = cpm.clearcoatRoughnessTexmapOn
				vrm.texmap_coat_ior_on = cpm.clearcoatIorTexmapOn
				vrm.texmap_coat_color_on = cpm.clearcoatAbsorptionTexmapOn
				vrm.texmap_coat_bump_on = cpm.clearcoatBumpTexmapOn
				vrm.texmap_sheen_on = cpm.sheenColorTexmapOn
				vrm.texmap_sheen_glossiness_on = cpm.sheenRoughnessTexmapOn
				vrm.texmap_environment_on = cpm.bgOverrideReflectTexmapOn or cpm.bgOverrideRefractTexmapOn

				--V-Ray material output
				vrm
			),			
			
		fn gt_createVRayMtl orig_mtl = 
			(

				undo off
				(
					

				-- check if this is a material; if not, just return the value
				if ((superclassof orig_mtl)!=material) then 
					return orig_mtl

				-- convert sub-textures of this material
				local sub_texs_count=getNumSubTexmaps orig_mtl
				
				if sub_texs_count>0 then 
					(
						local sub_tex = undefined
						
						for i=1 to sub_texs_count do 
							(
								sub_tex = getSubTexmap orig_mtl i
								if(sub_tex != undefined) then
									setSubTexmap orig_mtl i (gt_createVRayTexture sub_tex)
							)
					)

				-- check if we have already converted this material; if yes, just return
				-- the already converted one
				local i=findItem glsc_orig_mtls orig_mtl

				if i>0 then 
					return glsc_new_mtls[i]

				-- recurse into sub-materials, if any
				local sub_count=getNumSubMtls orig_mtl
				if sub_count>0 then (
					local sub_mtl = undefined
					for i=1 to sub_count do (
						sub_mtl=getSubMtl orig_mtl i
						if(sub_mtl != undefined) then
							setSubMtl orig_mtl i (gt_createVRayMtl sub_mtl)
					)
				)
				
				-- get the material class id
				cls=classof orig_mtl

				-- if this is already a V-Ray compatible material, return the same one
				if (gt_isVRayCompatible cls) then 
				(
					--If the VRayMtl comes from old scenes with unlocked hilight glossiness
					if cls==VRayMtl and orig_mtl.reflection_lockGlossiness==off then (
						orig_mtl.hilight_glossiness = orig_mtl.reflection_glossiness
						orig_mtl.texmap_hilightGlossiness = undefined
						orig_mtl.reflection_lockGlossiness = on
					)
					return orig_mtl
				)

				
				local new_mtl=case cls of 
					(
						StandardMaterial: gt_convertFrom_Standard orig_mtl
						RaytraceMaterial: gt_convertFrom_Raytrace orig_mtl
						Architectural: gt_convertFrom_Architectural orig_mtl
						PhysicalMaterial: gt_convertFrom_PhysicalMaterial orig_mtl

				-- 				fR_Advanced: convertFrom_fR_Advanced orig_mtl
				-- 				fR_Metal: convertFrom_fR_Metal orig_mtl
				-- 				fR_Glass: convertFrom_fR_Glass orig_mtl

						Blend: gt_convertFromBlend orig_mtl
						Shellac: gt_convertFromShellac orig_mtl
						DoubleSided: gt_convertFromDoubleSided orig_mtl

						CoronaMtl: gt_convertFrom_CoronaMtl orig_mtl
						CoronaLayeredMtl: gt_convertFrom_CoronaLayeredMtl orig_mtl
						CoronaLightMtl: gt_convertFrom_CoronaLightMtl orig_mtl
						CoronaRaySwitchMtl: gt_convertFrom_CoronaRaySwitchMtl orig_mtl
						CoronaPortalMtl: VRayMtl name:orig_mtl.name
						CoronaVolumeMtl: gt_convertFrom_CoronaVolumeMtl orig_mtl
						CoronaHairMtl: gt_convertFrom_CoronaHairMtl orig_mtl
						CoronaSkinMtl: gt_convertFrom_CoronaSkinMtl orig_mtl
						CoronaPhysicalMtl: gt_convertFrom_CoronaPhysicalMtl orig_mtl "full"


						-- Do not convert Arch&Design materials - V-Ray supports them directly
						-- Arch___Design__mi: orig_mtl

						default: orig_mtl
					)
					
				-- if converted, add to the list of converted materials
				if new_mtl!=orig_mtl then (
					append glsc_orig_mtls orig_mtl
					append glsc_new_mtls new_mtl
					--here we create a node in Slate Material Editor
					--for newly created material
					gt_createSMENodeByRef orig_mtl new_mtl 
				)
				
				-- return the new material
				new_mtl

					)	-- undo off			
			),			
		
		-- convert node materials to VRay materials
		fn gt_mtlsToVRay node = 
			(
-- 				format "gt_mtlsToVRay: % \n" node
				
				if (isValidNode node) then (
					if debug do
						format " Converting node % materials...\n" node.name
					local m=gt_createVRayMtl node.material
					node.material=m
					
					if debug do
						format "Done\n"
				)
			),
	
		-- traverse scene nodes in order to convert specific properties
		-- like materials, shadows, ets., to VRay and replace the originals with the converted ones
		fn gt_replaceNodeProperties node = 
			(
				
				if isValidNode node then (
					gt_coronaLightTexToVRayTex node
					gt_mtlsToVRay node
					gt_shadowsToVRay node
				)
				
				for i=1 to node.children.count do
					(
						-- update progress info
						gui_CoronaToVrayConverter.lbl_info.text = "Converting Material " + (i as string) + " of " + (node.children.count as string)
						gui_CoronaToVrayConverter.prog_convert.value = 100. * i/node.children.count	
						
						gt_replaceNodeProperties node.children[i]
						
						gc()
					)
				gui_CoronaToVrayConverter.lbl_info.text = ""
				gui_CoronaToVrayConverter.prog_convert.value = 0
			),

				
		-- *********************************************************************** Cameras Functions ****************************************************************************
			
		fn gt_replaceCoronaCameraNode node = 
			(
				if debug do
					format "  Camera : % \n"  node.name
								
				local newCamera=CreateInstance VRayPhysicalCamera specify_fov: on pos: node.pos
				local newCameraName = "_VRayPhysicalCamera"

				if (isValidNode node and newCamera!=undefined) do (
					--ViewportDisplay -> Basic&Display
					newCamera.targeted = node.targeted
					if node.targeted then
						newCamera.target_distance = node.targetDistance
					case node.showCone of (
						0: newCamera.show_camera_cone = 1 --always
						1: newCamera.show_camera_cone = 0 --selected
						2: newCamera.show_camera_cone = 2 --never
					)
					newCamera.horizon_on = node.showHorizonLine

					--Sensor&Lens
					newCamera.specify_fov = (node.fovSource == 0)
					newCamera.fov = node.fov
					newCamera.focal_length = node.focalLength
					newCamera.zoom_factor = node.zoomFactor
					newCamera.ISO = node.ISO
					newCamera.f_number = node.fStop
					case node.shutterParam of (
						0:(
							newCamera.type = 0
							newCamera.shutter_speed = node.shutterSpeed
						)
						1:(
							newCamera.type = 1
							newCamera.shutter_angle = node.shutterAngle --in deg
							if node.enableCameraMb == on do
								newCamera.shutter_offset = node.shutterOffset
						)
						2:(
							newCamera.type = 2
							newCamera.latency = node.mblurDuration
						)
					)
					--Distortion
					if node.distortionType == 1 do (
						newCamera.distortion_type = 1
						newCamera.distortion = node.distortionAmount
					)
					if node.distortionType == 2 do (
						newCamer.distortion_type = 3
						newCamera.distortion_map = node.distortionMap
					)
					--DOF&MotionBlur
					newCamera.use_dof = node.enableDof
					newCamera.use_moblur = node.enableCameraMb

					--Bokeh effects
					if node.overrideBokeh == on then (
						if node.bokehType == 1 do (
							newCamera.use_blades = on
							newCamera.blades_number = node.bokehBladeCount
							newCamera.blades_rotation = node.bokehRotation
						)
						if node.bokehType == 2 do (
							newCamera.bitmap_aperture_texmap = node.bokehMap
							newCamera.bitmap_aperture_affectExposure = node.bokehMapAffectsExposure
						)
					)
					--Bokeh advanced effects
					newCamera.center_bias = node.bokehCenterBias
					newCamera.anisotropy = node.bokehAnisotropy
					newCamera.optical_vignetting = node.bokehVignetting

					--Tilt&Shift
					newCamera.lens_tilt_auto = node.autoVerticalTilt
					newCamera.lens_tilt = node.verticalTilt
					newCamera.lens_horTilt = node.horizontalTilt
					newCamera.vertical_shift = node.verticalShift
					newCamera.horizontal_shift = node.horizontalShift

					--Environment&Clipping
					newCamera.clip_on = node.enableClipping
					newCamera.clip_near = node.clippingNear
					newCamera.clip_far = node.clippingFar
					newCamera.environment_near = node.enviroNear
					newCamera.environment_far = node.enviroFar
					/*
					--Tone mapping
					if node.colorMappingOverride == on and node.enableToneMapping then 
						(
						if node.vignette > 0.0 do (
							newCamera.vignetting = on
							newCamera.vignetting_amount = node.vignette
						)
						newCamera.whiteBalance_preset = 0
						newCamera.whiteBalance = node.colorTint
						--node.colorTint
						--node.whiteBalance (temperature)
					)
					*/			
				)

				if (isValidNode node) do (
					instCount = InstanceMgr.GetInstances node &arrInst
-- 					for i=1 to instCount do
-- 						arrInst[i].name = (arrInst[i].name+newCameraName)
					replaceInstances node newCamera
				)

				newCamera!=undefined
			),

		fn gt_replaceCameraNode node = 
			(
				if debug do
					format "  Camera : % \n"  node.name
										
				local cameraClassStandard=#(Freecamera, Targetcamera)

				local newCamera
				local newCameraName
				
				if (isValidNode node) do 
				(
					local camIdx = findItem cameraClassStandard (classof node)
					if (camIdx!=0) do 
					(
						--Implementation

						newCamera=CreateInstance VRayPhysicalCamera specify_fov: on
						newCameraName = "_VRayPhysicalCamera"
						

						if (newCamera!=undefined) do (
							local newCameraClass = classof newCamera
							if (camIdx == 1) then (
								newCamera.targeted = off
								newCamera.target_distance = node.baseObject.targetDistance
							)
							else (
								newCamera.targeted = off
								newCamera.target_distance = node.targetDistance
								newCamera.targeted = on
							)

							if (node.fov.isAnimated) then(
								classOfNodeController = classof node.fov.controller
								--newCamera.fov.controller=bezier_float()
								newCamera.fov.controller = createInstance classOfNodeController
								nkeys = node.fov.keys
								for k in nkeys do appendKey newCamera.fov.keys k
								for i=1 to nkeys.count do (
									newCamera.fov.keys[i].value=node.fov.keys[i].value
								)
							) else (
								newCamera.fov = node.fov -- By some reason Max Physical Camera does not accept fov value
							)

							newCamera.specify_fov = off

							if (node.showRanges == on) then (
								newCamera.environment_near = node.nearrange
								newCamera.environment_far = node.farrange
							)

							newCamera.clip_on = node.clipManually
							newCamera.clip_near = node.nearclip
							newCamera.clip_far = node.farclip

							if (node.mpassEnabled == on) then (
								local mpEffectClass = classof node.mpassEffect
								case mpEffectClass of (
									Depth_of_FieldMPassCamEffect: (
										newCamera.use_dof = on
										if (newCameraClass == Physical) then (
											newCamera.motion_blur_enabled = off
											newCamera.focus_distance = node.mpassEffect.focalDepth
											if (node.mpassEffect.useTargetDistance == on) then newCamera.specify_focus = 0
											else newCamera.specify_focus = 1
										)
										else newCamera.use_moblur = off
									)
									Motion_BlurMPassCamEffect: (
										newCamera.use_dof = off
										if (newCameraClass == Physical) do newCamera.motion_blur_enabled = on
										if (newCameraClass == VRayPhysicalCamera) do newCamera.use_moblur = on
									)
									Depth_of_Field__mental_ray: (
										newCamera.use_dof = on
									)
								)
							)

							instCount = InstanceMgr.GetInstances node &arrInst

							replaceInstances node newCamera
						)
					)
				)

				if (newCamera!=undefined) do (
					if debug do
						print (node.name)
				)

				newCamera!=undefined
			),

		fn gt_camerasToVRay selectedOnly = 
			(
				if debug do
					format " ********** Cameras Conversion ********** \n\n"
				
				local oldCameras=#()

				if debug do
					if selectedOnly do format "Converting scene Cameras for selected cameras only...\n"
					
				for i=1 to $cameras.count do 
					(				
						-- update progress info
						gui_CoronaToVrayConverter.lbl_info.text = "Converting Camera " + (i as string) + " of " + ($cameras.count as string)
						gui_CoronaToVrayConverter.prog_convert.value = 100. * i/$cameras.count	
							
						if selectedOnly then (
							if $cameras[i].isSelected do (
								if (classof $cameras[i] == CoronaCam) then (
									if (gt_replaceCoronaCameraNode $cameras[i]) do ( append oldCameras $cameras[i] )
								) else (
									if (gt_replaceCameraNode $cameras[i]) do ( append oldCameras $cameras[i] )
								)
							)
						) else (
							if (classof $cameras[i] == CoronaCam) then (
								if (gt_replaceCoronaCameraNode $cameras[i]) do ( append oldCameras $cameras[i] )
							) else (
								if (gt_replaceCameraNode $cameras[i]) do ( append oldCameras $cameras[i] )
							)
						)
					)
					
					gui_CoronaToVrayConverter.lbl_info.text = ""
					gui_CoronaToVrayConverter.prog_convert.value = 0
					
					if debug do
						format " ********** Finished Cameras Conversion ********** \n\n"
			),
	
	
		-- *********************************************************************** Proxies Functions ****************************************************************************

		-- Function to get the frame range of an animated proxy, and some other data if needed
		-- the corona proxies have a header with some info, with this we can read it
		-- there is probably a much beter way but this is how I got it working...
		-- got the idea from https://forums.cgsociety.org/t/reading-a-max-file-as-a-binary-file/1406772
		fn gt_getProxyInfo proxy debug:true =
			(
				aniStart = 0
				aniEnd = 0

				if classof proxy == CProxy then
				(
					if proxy.fileName != undefined then
					(		
						-- open the file as binary
						thebytestream = fopen proxy.fileName "rb"
						thesize = getfilesize proxy.fileName 
						thestring = ""
						
						-- go trough all the bytes
						i = 1
						while (i < thesize) do
							(
								escapeEnable = true						
								if keyboard.escPressed do 
								(
									fclose thebytestream
									exit	
								)
								
								thebyte = bit.intaschar (readbyte thebytestream)	
								-- add them all to a string for later processing
								if thebyte!= "" do
									thestring += thebyte
								
								-- if we find the word "Compressed" exit since thats the end of were our wanted info is, at least as of 2023-01-25
								if (findstring thestring "Compressed:" != undefined) do
								(
									exit
								)
								
								i+=1
								
								-- fail safe exit
								if i > 1000 do exit
							)
						
						fclose thebytestream

						-- this prints the header info, as of 2023-01-25 is like:
						/*
						C?
						Chaos Corona Proxy file.
						This file contains an animated Mesh object.
						This file is BINARY.
						Version: 8
						Date created: 2023-01-24 11:57:41
						Number of faces: 335872
						Number of vertices: 337318
						Number of mapping channels: 2
						Number of normals: 403639
						Number of hair: 0
						Bounding box size (original units) : 209.084259 223.571426 354.500092
						Bounding box size (millimeters) : 5310.74023 5678.71436 9004.30176
						Original material name: MT_PM_Populus_x_canadensis_01
						Original object name: MT_PM_Populus_x_canadensis_01_03
						Animation snapshot count: 1
						Time per animation snapshot [microseconds]: 33333
						Animation time start: 0
						Scale: 0.0393700786
						Compressed: 	
						*/
						if debug then (
						format "% \n" thestring
						format "\n"	
						)					
						-- make it into a stringstream so I can get by line
						strHeader = thestring as stringStream
						
						while not eof strHeader do 
						(
							escapeEnable = true						
							if keyboard.escPressed do exit	
							
							str = readline strHeader
							
							strLine = filterString str ":"
							
							-- if needed we can get other info here, for now just the animation length
							if (strLine[1] == "Animation time start") then (
								aniStart = strLine[2] as integer
							)	
							
							if (strLine[1] == "Animation snapshot count") then (
								aniEnd = strLine[2] as integer
							)	
						)
						
						close strHeader
					)
					
				)
				return #(aniStart, aniEnd)
			),
			
		--Convert Corona Proxy to VRay Proxy
		fn gt_getOrigCoronaProxy  selectedOnly =  --get list of original proxies
			( 
				arrayOrigCoronaProxy = #()
				arrayOrigCoronaProxy_pathes = #()
				arrayUniqueCoronaPeoxy = #()  -- these are proxies that are not instances but have the same proxy file on disk
				
					for i in objects where (classOf i == CProxy)  and (not selectedOnly or selectedOnly and i.isSelected) do
					(
						if (findItem arrayOrigCoronaProxy_pathes  i.filename) == 0 and doesFileExist i.filename != false then
						(
						append arrayOrigCoronaProxy i
						append arrayOrigCoronaProxy_pathes i.filename
					)else (
							append arrayUniqueCoronaPeoxy i
						)
				)

				return #(arrayOrigCoronaProxy, arrayUniqueCoronaPeoxy)
			),

		fn gt_getData_InstProxy_byOrig ocp =  --get list of dependent proxies
			(
				arrayCoronaProxy = #()
				for i in objects where classOf i == CProxy do
				(
					if i.filename != undefined and i.filename == ocp.filename do 
						(
							aniLength = gt_getProxyInfo i debug:false
							append arrayCoronaProxy #(i, i.name, getFilenameFile  i.filename, getFilenamePath i.filename, i.transform, i.previzType, i.material, aniLength)	
						)
				)
				arrayCoronaProxy
			),
				
		fn gt_VrayProxy_FromExistingProxyFromDisk thePath CrnProxy =
			(
				aniLength = gt_getProxyInfo CrnProxy debug:false						
					
				theProxy = VRayProxy()
				theProxy.name = CrnProxy.name
				theProxy.wirecolor = CrnProxy.wirecolor
				theProxy.gbufferChannel = CrnProxy.gbufferChannel
					
				theLayer = LayerManager.getLayerFromName CrnProxy.layer.name
					if theLayer != undefined then
						theLayer.addNode theProxy
				
				theCompletePath = ""
				theFileName = getFilenameFile CrnProxy.filename
				
				if thePath != undefined  then (
					theCompletePath = thePath + "\\" + theFileName + ".vrmesh"
				)else(
					thePath = getFilenamePath CrnProxy.filename
					theCompletePath = thePath + theFileName + ".vrmesh"
				)
				
				theProxy.fileName = theCompletePath
				theProxy.transform = CrnProxy.transform						
				theProxy.material = CrnProxy.material
				
				-- if it was an animation set the parameters
				if aniLength[2] > 1 then -- has animation
					(
						-- Vray 	0: loop, 		1: Play Once, 	2: PingPong , 				3: Still
						-- Corona 	0: Repeat, 	1: PingPong, 	2: Freez out of range, 	3: hide out of range
						if CrnProxy.animationMode == 0 then
							theProxy.anim_type = 0
						else if CrnProxy.animationMode == 1 then
							theProxy.anim_type = 2
						else
							theProxy.anim_type = 0
						
						theProxy.anim_offset = CrnProxy.animationOffset
						theProxy.anim_speed = CrnProxy.animationSpeedMultiplier 
					)
					
				return theProxy
			),
		
		fn gt_coronaProxiesToVRay thePath exportAnimation selectedOnly useExistingProxy:false debug:false = 
			(
				undo off
				(
				local saveAniStart = animationRange.start
				local saveAniEnd = animationRange.end

				-- Get all proxies
				AllProxies = gt_getOrigCoronaProxy selectedOnly
				OrigCorPr = AllProxies[1]
				UniqueCorPr = AllProxies[2]

				if debug then (
					format "\n ********** Proxy Conversion ********** \n"
					format "Processing % Instanced Proxies and % unique Proxies \n\n" OrigCorPr.count UniqueCorPr.count
				)

				local proxiesToDelete = #()
				local loopProxies = #()
				
				-- Process Original Proxies
				for i = 1 to OrigCorPr.count do 
						(
							-- update progress info
							gui_CoronaToVrayConverter.lbl_info.text = "Converting Instanced Proxy " + (i as string) + " of " + (OrigCorPr.count as string)
							gui_CoronaToVrayConverter.prog_convert.value = 100. * i/OrigCorPr.count	
							
							-- get all needed data from each proxy
							theData_InstCorPr = gt_getData_InstProxy_byOrig OrigCorPr[i]
								
							if debug then (
								format " Proxy % : % \n" i OrigCorPr[i].name
								
									for d = 1 to theData_InstCorPr[1].count do
										format "        Data: % % \n" d theData_InstCorPr[1][d]
								)								
							
							-- set the filename, minus extension
							theCompleteFileName = ""
							
							if thePath != undefined  then 
								(
									theCompleteFileName = thePath + "\\" + theData_InstCorPr[1][3]
								) else 
									(
										theCompleteFileName = theData_InstCorPr[1][4] + theData_InstCorPr[1][3]
									)

							if debug then (
								for p = 1 to theData_InstCorPr.count do 
										format "         Instance % of %: % \n" p theData_InstCorPr.count theData_InstCorPr[p][1].name	
									
								format "            Saving the Proxy to: % \n" (theCompleteFileName + ".vrmesh")
								)

								
							-- check if the file exsits and if the user wants to use the existing file
							if ((doesFileExist (theCompleteFileName + ".vrmesh") ) and (useExistingProxy) )then 
									(
										if debug then
											format "             *** Using existing proxy from : % \n\n" (theCompleteFileName + ".vrmesh")
										
										local tempProxy = gt_VrayProxy_FromExistingProxyFromDisk thePath OrigCorPr[i]
										
										append proxiesToDelete OrigCorPr[i]

										for d = 1 to theData_InstCorPr.count do
										(
-- 											format "     Data: % % \n" d theData_InstCorPr[d]
											
											-- 1: proxy , 
											-- 2: object name 
											-- 3: proxy filename without extension  
											-- 4: proxy file path
											-- 5: transform
											-- 6: display type
											-- 7: material
											-- 8: animated frame range
											
											instanceReplace OrigCorPr[i] tempProxy
										)
										
									) else 
										(
											if debug then
												format "             Exporting Proxy : % \n" OrigCorPr[i]
										
											tempObj = instance OrigCorPr[i] wirecolor:red
											tempObj.transform = (matrix3 [1,0,0] [0,1,0] [0,0,1] [0,0,0])
											tempObj.previzType = 3
											editMod = Edit_Mesh()
											editMod.name = "editMod"
											addModifier tempObj editMod
											objNF = (tempObj.numfaces/10.0) as integer -- preview 1/10 ot numfaces
												
											for m = 1 to tempObj.modifiers.count do (
												if tempObj.modifiers[m].name == "editMod" do
													deleteModifier tempObj m
											)

											tempObj.name = theData_InstCorPr[1][3]
												
											-- has animation ?
											local startFrame = 0
											local endFrame = 0
											if theData_InstCorPr[1][8][2] > 1 then -- has animation
													(
														startFrame = theData_InstCorPr[1][8][1]
														endFrame = theData_InstCorPr[1][8][2]
														animationRange = interval startFrame endFrame
													)
										
											local tempVRayPr = undefined
											-- Export the vrayproxy
											select tempObj

											if theData_InstCorPr[1][8][2] > 1 and exportAnimation then -- has animation
												(	
													tempVRayPr = vrayMeshExport tempObj autoCreateProxies:true animation:true animationRange:#explicit animationStart:startFrame  animationEnd:endFrame animationRedrawViews:false maxPreviewFaces:objNF  meshfile:theCompleteFileName
												) else (
													tempVRayPr = vrayMeshExport tempObj autoCreateProxies:true maxPreviewFaces:objNF  meshfile:theCompleteFileName
												)

											-- if it was an animation set the parameters
											if theData_InstCorPr[1][8][2] > 1 then -- has animation
												(
													-- Vray 	0: loop, 		1: Play Once, 	2: PingPong , 				3: Still
													-- Corona 	0: Repeat, 	1: PingPong, 	2: Freez out of range, 	3: hide out of range
													if OrigCorPr[i].animationMode == 0 then
														tempVRayPr[1].anim_type = 0
													else if OrigCorPr[i].animationMode == 1 then
														tempVRayPr[1].anim_type = 2
													else
														tempVRayPr[1].anim_type = 0
													
													tempVRayPr[1].anim_offset = OrigCorPr[i].animationOffset
													tempVRayPr[1].anim_speed = OrigCorPr[i].animationSpeedMultiplier 
												)
											
											tempVRayPr[1].gbufferChannel = OrigCorPr[i].gbufferChannel
											tempVRayPr[1].display = 1
											tempVRayPr[1].visibility_lists_type = 0
											tempVRayPr[1].visibility_list_names[1] = false
											for i in (gt_getData_InstProxy_byOrig OrigCorPr[i]) do i[1].name = uniqueName tempObj.name
												
												
-- 											for d = 1 to theData_InstCorPr.count do
-- 											(
												-- 1: proxy , 
												-- 2: object name 
												-- 3: proxy filename without extension  
												-- 4: proxy file path
												-- 5: transform
												-- 6: display type
												-- 7: material
												-- 8: animated frame range
-- 											)

											--replace each corona proxy with vray proxy -> tempVRayPr
											instanceReplace OrigCorPr[i] tempVRayPr[1]		
											delete tempVRayPr

										)

						gui_CoronaToVrayConverter.lbl_info.text = ""
						gui_CoronaToVrayConverter.prog_convert.value = 0
					) -- end OrigCorPr.count			

				-- I don't know why some proxies don't show up unless you select it and change the display type , Bug ?
				-- so here I ran trought the proxies and set the display mode again
				for o in objects where (classOf o == VRayProxy) do 
					o.display = 1
					
				if debug then
					format "\n"

				-- remove any vrayProxies from the unique array. since some of the proxies in that array are now vrayproxy as even do they are not instances some might have the same proxy file which we replaced above
				UniqueCorPrClean = #()
				for i = 1 to UniqueCorPr.count do 
					if (classof UniqueCorPr[i] == CProxy ) do 
							append UniqueCorPrClean UniqueCorPr[i]

				-- these proxies are the ones that have the same proxy file but are not instances of some other proxy.
				-- so we manually create a vray proxy and fill up all the info
				-- process Unique proxies
						for i = 1 to UniqueCorPrClean.count do 
							(		
								if debug then
									format "   Unique Proxy : % \n" UniqueCorPrClean[i]
								-- update progress info
								gui_CoronaToVrayConverter.lbl_info.text = "Converting Iunique Proxy " + (i as string) + " of " + (UniqueCorPr.count as string)
								gui_CoronaToVrayConverter.prog_convert.value = 100. * i/UniqueCorPr.count	
								
								local theProxy = gt_VrayProxy_FromExistingProxyFromDisk thePath UniqueCorPrClean[i]	
								theProxy.display = 0
								theProxy.display = 1
							)	
						gui_CoronaToVrayConverter.lbl_info.text = ""
						gui_CoronaToVrayConverter.prog_convert.value = 0						

				if proxiesToDelete.count !=0 do
					delete proxiesToDelete
			
				if UniqueCorPrClean.count != 0 do 
					delete UniqueCorPrClean

				animationRange = (interval saveAniStart saveAniEnd)

				if debug then (
					format "\n ********** Finished Proxy Conversion ********** \n\n"
				)
			) -- undo off
			),	
							
	
		-- *********************************************************************** Enviorment Function  ****************************************************************************
			
		fn gt_restoreEnviormentMaps =
			(
					if debug do
						format "\nRestoring Environment Maps from Corona\n"	
					
					
					if isCoronaCurrent() then
					(
						if theCoronaSingleMap != undefined then
							renderers.current.environment_gi_map = theCoronaSingleMap
						renderers.current.environment_gi_color = theCoronaSingleColor
						renderers.current.environment_gi_on = theCoronaSingleEnabled
						renderers.current.environment_gi_map_on = theCoronaSingleEnabledTex

						-- reflection map
						if theCoronaReflectOverride != undefined then 
							renderers.current.environment_rr_map = theCoronaReflectOverride
						renderers.current.environment_rr_color = theCoronaReflectColor 
						renderers.current.environment_rr_on = theCoronaReflectEnabled 
						renderers.current.environment_rr_map_on = theCoronaReflectEnabledTex 

						-- refraction map
						if theCoronaRefractOverride != undefined then 
							renderers.current.environment_refract_map = theCoronaRefractOverride
						renderers.current.environment_refract_color = theCoronaRefractColor 
						renderers.current.environment_refract_on = theCoronaRefractEnabled 
						renderers.current.environment_refract_map_on = theCoronaRefractEnabledTex
					)
	
			),
			
		-- *********************************************************************** Main Convert Function  ****************************************************************************
			
		fn gt_ConvertToVRay selectedOnly =
			(
				glsc_orig_mtls=#()
				glsc_new_mtls=#()
				glsc_orig_texmaps=#()
				glsc_new_texmaps=#()

				-- Disable scene redraw during the conversion in order to prevent 3ds Max from updating the viewports after every change
				with redraw off
				(
					if gui_CoronaToVrayConverter.chk_materials.checked then
					(
						if selectedOnly then 
							(
								local numNodes = selection.count
								for iNode=1 to numNodes do (
									if debug do
										print selection[iNode]
									gt_replaceNodeProperties selection[iNode]
								)
							) else (
								gt_replaceNodeProperties rootNode

								if debug do
									format "Converting Material Editor materials...\n"
								
								for i=1 to 24 do (
									local m=gt_createVRayMtl meditmaterials[i]
									meditmaterials[i]=m
								)
								if debug do
									format "Done\n"
								
								if debug do
									format "Converting Environment Map...\n"
								
								-- Enviorment Map
								local envtex=gt_createVRayTexture environmentMap
								if (envtex!=undefined) do (
									gt_createSMENodeByRef environmentMap envtex deleteOldNode:deleteOldNodes
									environmentMap=envtex
								)
								theEnviormentColor = backgroundColor
								
								-- Corona Enviorments

								-- Single Map
								if renderers.current.bg_texmap != undefined then (
									local envtex = gt_createVRayTexture renderers.current.bg_texmap
									if (envtex!=undefined) do (
										gt_createSMENodeByRef renderers.current.bg_texmap envtex 
										renderers.current.bg_texmap = envtex
										theCoronaSingleMap = envtex	
									)
								)
					
								theCoronaSingleColor = renderers.current.bg_color
								if (renderers.current.bg_source == 1) then
									theCoronaSingleEnabled = true
								else
									theCoronaSingleEnabled = false
								theCoronaSingleEnabledTex = renderers.current.bg_texmapUseMain
								
								-- Direct Visibility Override
								if renderers.current.bg_texmapDirect != undefined then (
									local envtex = gt_createVRayTexture renderers.current.bg_texmapDirect
									if (envtex!=undefined) do (
										gt_createSMENodeByRef renderers.current.bg_texmapDirect envtex 
										renderers.current.bg_texmapDirect = envtex
										theCoronaVisibilityOverride = envtex	
									)
								)
								theCoronaVisibilityColor = renderers.current.bg_colorDirect
								theCoronaVisibilityEnabled = renderers.current.bg_overrideDirect 
								theCoronaVisibilityEnabledTex = renderers.current. bg_texmapUseDirect
								
								-- Reflection Override
								if renderers.current.bg_texmapReflect != undefined then (
									local envtex = gt_createVRayTexture renderers.current.bg_texmapReflect
									if (envtex!=undefined) do (
										gt_createSMENodeByRef renderers.current.bg_texmapReflect envtex 
										renderers.current.bg_texmapReflect = envtex
										theCoronaReflectOverride = envtex	
									)
								)
								theCoronaReflectColor = renderers.current.bg_colorReflect
								theCoronaReflectEnabled = renderers.current.bg_overrideReflect
								theCoronaReflectEnabledTex = renderers.current.bg_texmapUseReflect
								
								-- Refraction Override
								if renderers.current.bg_texmapRefract != undefined then (
									local envtex = gt_createVRayTexture renderers.current.bg_texmapRefract
									if (envtex!=undefined) do (
										gt_createSMENodeByRef renderers.current.bg_texmapRefract envtex 
										renderers.current.bg_texmapRefract = envtex
										theCoronaRefractOverride = envtex	
									)
								)	 
								theCoronaRefractColor = renderers.current.bg_colorRefract
								theCoronaRefractEnabled = renderers.current.bg_overrideRefract 
								theCoronaRefractEnabledTex = renderers.current.bg_texmapUseRefract
								
								
								if debug do
									format "Done\n"

						)					
					)

					
					if gui_CoronaToVrayConverter.chk_lights.checked then
						gt_lightsToVRay selectedOnly
					
					if gui_CoronaToVrayConverter.chk_cameras.checked then
						gt_camerasToVRay selectedOnly
					
					if gui_CoronaToVrayConverter.chk_proxy.checked then
					(
						if gui_CoronaToVrayConverter.chk_custom_proxyPath.checked then (
							if proxyCustomPath != undefined then
								gt_coronaProxiesToVRay proxyCustomPath gui_CoronaToVrayConverter.chk_exportAnimation.checked selectedOnly useExistingProxy:gui_CoronaToVrayConverter.chk_useExisting.checked debug:gui_CoronaToVrayConverter.chk_debug.checked
							else
								messageBox "Proxy Custom Path not defined\nNot converting proxies"
						)
						else
							gt_coronaProxiesToVRay undefined gui_CoronaToVrayConverter.chk_exportAnimation.checked selectedOnly useExistingProxy:gui_CoronaToVrayConverter.chk_useExisting.checked debug:gui_CoronaToVrayConverter.chk_debug.checked
					)

				)

				gui_CoronaToVrayConverter.lbl_info.text = "Finished Converting Scene"
				
				glsc_orig_mtls=#()
				glsc_new_mtls=#()
				glsc_orig_texmaps=#()
				glsc_new_texmaps=#()
			)			
					
		-- *********************************************************************************************************************************************************************

			
	)
	
	Local CTOV =  Convert_Scene_CoronatoVray()
	
	
	rollout gui_CoronaToVrayConverter "Corona To Vray Converter" width:500 height:395
		(
			
			checkbox chk_materials "Materials" width:60 height:15 checked:true align:#left offset:[0,20] 
			checkbox chk_2sideMat "Translucency: Convert To VRay2SidedMtl" width:220 height:15 checked:false offset:[70,0] tooltip:"When the script finds the translucency slot is used in a corona material it will create a Vray2Side Material stead of a VrayMtl"
			checkbox chk_CrnColorCorrect "Convert Corona Color Correct" width:220 height:15 checked:false offset:[70,0] tooltip:"Vray 6 supports Corona Color Correct but if you need to you can still convert it.\nKeep in mind the result will not be exactly the same if you convert it\nFor GPU you do need to convert it"
			GroupBox grp_MatOptions "Options:" pos:[75,25] width:415 height:60 align:#left
			
			checkbox chk_lights "Lights" width:80 height:15 checked:true align:#left offset:[0,0] 
			checkbox chk_cameras "Cameras" width:80 height:15 checked:true align:#left offset:[0,0] 
			
			checkbox chk_proxy "Proxies" width:60 height:15 checked:true offset:[0,0] align:#left across:3
			checkbox chk_custom_proxyPath "Custom Path:" width:80 height:20 checked:false offset:[-90,15] tooltip:"Save the new vrayproxy files to a custom folder. If disabled they will be placed where the corona proxy was"
			button btn_proxy_path "Proxy Path ..."  width:320 height:20 enabled:false offset:[-80,15]
			checkbox chk_exportAnimation "Export Animation (SLOW...)" width:150 height:20 checked:true offset:[68,-5] tooltip:"Export animated proxies, If disabled they will get export as static.\nKeep in mind that exporting animation is very slow, so be prepared to wait"
			checkbox chk_useExisting "Use Existing Proxy Files" width:150 height:20 checked:false offset:[68,-5] tooltip:"When re-converting the scene you might want to use the existing proxy files to save time"
			GroupBox grp_ProxyOptions "Options:" pos:[75,130] width:415 height:80 align:#left
			
			checkbox chk_bitmaps "Bitmaps to VrayBitmap" width:125 height:15 checked:true align:#left offset:[0,0] tooltip:"Converts all Bitmaps and Corona Bitmaps to VrayBitmap. When converting the materials some bitmaps are converted to VrayBitmap even if this is not checked."
			checkbox chk_elemets "Render Elements" width:200 height:15 checked:true align:#left offset:[0,0] tooltip:"Try to convert equivalent Render Elements"
			
			GroupBox grp_WhatToConvert "What to Convert:" pos:[5,5] width:490 height:255 align:#left
			

			radiobuttons rdo_onWhat "" labels:#("Scene", "Selection") offset:[-30,25] align:#right
			button btn_convert "Start" width:200 height:50 offset:[0,0] align:#right
			GroupBox grp_WorkON "Work On:" pos:[275,270] width:220 height:95 align:#right
			
			checkbox chk_debug "Debug" width:200 height:15 checked:true align:#left offset:[0,-80] tooltip:"Prints some information about whats happening to the Listener"
			checkbox chk_cleanNodes "Clean Unused Nodes" width:200 height:15 checked:false align:#left offset:[0, 0] tooltip:"Delete unsed converted nodes in the slate editor"
			colorpicker clr_UnsupportedColor  "Unsupported Maps Color" color:[255,0,0] modal:false  offset:[0, 0] tooltip:"Unsuported maps will get  vray color map with this color"
			
			label lbl_info "info: ..........." width:260 offset:[-110, 5]
			progressbar prog_convert width:480 height:9 align:#left  offset:[0, 0]
			
			on gui_CoronaToVrayConverter open  do
					(
					gui_CoronaToVrayConverter.title = "Corona To Vray Scene Converter        v" + CTOV.version
					)
		
			on chk_materials changed state do
				(
					chk_2sideMat.enabled = State
					
				)
			
			on chk_custom_proxyPath changed state do
				(
					btn_proxy_path.enabled = state
				)
				
			on chk_proxy changed state do
				(
					chk_custom_proxyPath.enabled = state
					if chk_custom_proxyPath.checked do
						btn_proxy_path.enabled = state
					chk_exportAnimation.enabled = state
				)
			
			on btn_proxy_path pressed do
			(
				CTOV.proxyCustomPath = getSavePath "Select Save Folder" initialDir:(maxFilePath)
				
				if CTOV.proxyCustomPath != undefined then 
					(
-- 						format "path:% \n" proxyCustomPath
						-- if the path is too long to display on the button then display just 75 characters in the button and the complete path in the tooltip
						newoutput= substring CTOV.proxyCustomPath 1 75
						if CTOV.proxyCustomPath.count > 75 then newoutput +="..."
						btn_proxy_path.caption = newoutput
						btn_proxy_path.tooltip = CTOV.proxyCustomPath
					)	
			)
			
			on btn_convert pressed do
				(
					start = timeStamp()
					
					clearListener()
					
					-- Convert Materials , Lights, Cmaeras and Proxies
					if rdo_onWhat.state == 1 then
						CTOV.gt_ConvertToVRay false -- Scene
					else 
						CTOV.gt_ConvertToVRay true -- Selection

					-- Convert Bitmaps
					if chk_bitmaps.checked do 
						(
							CTOV.replaceImgCounter=0
							CTOV.replaceImgNums = (getclassinstances Bitmaptexture processChildren:true).count
						
							CTOV.gt_VRayBitmap2VRayBitmap()
						)
					
					-- Convert render Elements						
					if chk_elemets.checked then
						CTOV.convert_RenderElements_from_CoronaToVray deleteOld:true debug:CTOV.debug
						
					-- checks to see if the current render is vray if not set it as default
					if CTOV.isVRayCurrent() == false do renderers.current = vray()
					
					-- restore eviorment Maps from corona
					if chk_materials.checked then
						CTOV.gt_restoreEnviormentMaps()
					
					-- clean slate editor unused nodes
					if chk_cleanNodes.checked do
						CTOV.removeUnsuedNodesfromSlateEditor debug:chk_debug.checked
					
					end = timeStamp()
					txtStr = "Finished. Conversion took " + (((end - start) / 1000.0) as string)  + " seconds" 
					lbl_info.text =  txtStr
				)
				
			on chk_debug changed state do 
				CTOV.debug = state
			
			
			on clr_UnsupportedColor changed arg do 
			(
				CTOV.unsupportedColor = arg
			)
			
		)
		
	createdialog gui_CoronaToVrayConverter
)